/*
 * Java
 *
 * Copyright 2013-2021 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.kf;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import ej.kf.AlreadyLoadedFeatureException;
import ej.kf.Feature;
import ej.kf.InvalidFormatException;
import ej.kf.Kernel;

/**
 * Default loader: load a static Feature from its name.
 */
public class StaticFeatureLoader implements IFeatureLoader{

	private static final int MAX_FEATURE_NAME = 256;

	private static final int ERROR_FORMAT_MAX_FEATURE_NAME = InvalidFormatException.FIRST_SPECIFIC_ERROR;
	private static final int ERROR_FORMAT_STATIC_FEATURE_NOT_FOUND = InvalidFormatException.FIRST_SPECIFIC_ERROR+1;

	/**
	 * Always return true.
	 */
	@Override
	public boolean canLoad(byte[] signature) {
		return true;
	}

	@Override
	public boolean canUnload() {
		return true;
	}

	@Override
	public Feature load(IKernelFeatureLoader kernelLoader, byte[] signature, InputStream is) throws IOException, AlreadyLoadedFeatureException, InvalidFormatException{
		InputStreamReader r = new InputStreamReader(is);
		char[] chars = new char[MAX_FEATURE_NAME+1];
		int signatureLength = signature.length;
		int totalNbRead = 0;
		for(int i=-1; ++i<signatureLength;){
			char c = (char)(signature[i]&0xff);
			if(c != 0){
				chars[totalNbRead++] = c;
			}
			else{
				// name is lower than 4 bytes
				break;
			}
		}

		int followingNbRead = r.read(chars, totalNbRead, MAX_FEATURE_NAME-totalNbRead);
		if(followingNbRead != -1) {
			totalNbRead += followingNbRead;
		}

		if(totalNbRead>MAX_FEATURE_NAME) {
			throw new InvalidFormatException(ERROR_FORMAT_MAX_FEATURE_NAME);
		}

		long featureSectionDescriptor = KernelNatives.loadStaticFeatureByName(chars, totalNbRead);

		if(featureSectionDescriptor>=0){
			// successfully loaded
			return new Feature(this, featureSectionDescriptor);
		}
		else{
			// error
			switch((int)featureSectionDescriptor){
			case KernelNativesConstants.ERROR_LOADSTATIC_ALREADYLOADED:{
				// Get back the Feature
				Feature f = Kernel.getFeatureByName(new String(chars, 0, totalNbRead));
				assert (f != null);
				throw new AlreadyLoadedFeatureException(f);
			}
			case KernelNativesConstants.ERROR_LOADSTATIC_NOT_FOUND: throw new InvalidFormatException(ERROR_FORMAT_STATIC_FEATURE_NOT_FOUND);
			default: throw new RuntimeException();
			}
		}
	}

	@Override
	public boolean unload(Feature f) {
		// nothing to do, the static features are linked to the Kernel.
		// They are considered to be successfully uninstalled for KF lifecycle point of view.
		KernelNatives.unlinkToNative(f);
		return true;
	}

}
