/*
 * Java
 *
 * Copyright 2011-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.sp.remote;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;

import com.is2t.sp.remote.request.ReqInterruptRequest;
import com.is2t.sp.remote.request.SPRequest;

import ej.sp.EmptyBlockException;

public class RemoteShieldedPlug {

	private final DataOutputStream dos;

	private final RequestReader requestReader;

	private final Hashtable idToRequest;
	private char nextRequestId;

	public RemoteShieldedPlug(RemoteShieldedPlugConnector connector) {
		InputStream is = connector.getInputStream();
		OutputStream os = connector.getOutputStream();
		this.dos = new DataOutputStream(os);
		requestReader = new RequestReader(new DataInputStream(is), this);

		connector.connect(requestReader);
		this.idToRequest = new Hashtable();
	}

	protected synchronized SPRequest removeRequest(int id){
		return (SPRequest) idToRequest.remove(new Integer(id));
	}

	public Object executeInterruptableRequest(SPRequest request) throws InterruptedException{
		try{
			return executeRequest(request, true);
		}
		catch(EmptyBlockException e){
			//should not occur
			throw new RemoteSPError(e);
		}
	}

	public Object executeReadBlockRequest(SPRequest request) throws EmptyBlockException{
		try{
			return executeRequest(request, false);
		}
		catch(InterruptedException e){
			//should not occur
			throw new RemoteSPError(e);
		}
	}

	public Object executeRequest(SPRequest request) {
		try{
			return executeRequest(request, false);
		}
		catch(InterruptedException e){
			//should not occur
			throw new RemoteSPError(e);
		}
		catch(EmptyBlockException e){
			//should not occur
			throw new RemoteSPError(e);
		}
	}

	private Object executeRequest(SPRequest request, boolean interruptible) throws InterruptedException, EmptyBlockException{
		//add the request in the pool and send it in the output stream
		char requestId;
		synchronized (this) {
			//get a free id
			while(idToRequest.containsKey(new Integer(requestId = nextRequestId++))) {

			}

			idToRequest.put(new Integer(requestId), request);
			sendRequest(request, requestId);
		}
		// Store the interrupted exception if we catch one. Null if no InterruptedException has been caugth.
		InterruptedException interruptedException = null;
		boolean stopWaitingForReply = false;
		//wait for a response
		synchronized (request) {
			while(!request.hasResponse() && !stopWaitingForReply){
				try{
					request.wait();
				}
				catch(InterruptedException e){
					interruptedException = e;
					if(interruptible){
						// Stop the wait for response loop
						stopWaitingForReply = true;
					}
					// else: method must no be interrupted, continue to wait
				}
			}
		}

		// Check if the thread has been interrupted while waiting for the reply
		if(interruptedException != null){
			//check if this request is interruptible
			if(interruptible){
				// WARNING: this code must be executed outside of the synchronized (request) to avoid deadlock
				executeRequest(new ReqInterruptRequest(requestId));
				removeRequest(requestId);
				throw interruptedException;
			}
			else {
				//set interrupt flag
				Thread.currentThread().interrupt();
			}
		}

		//return the response
		return request.getReplyResult();
	}

	private void sendRequest(SPRequest request, int requestId){
		try{
			dos.writeChar(requestId);
			dos.writeByte(request.commandID);
			dos.writeShort(request.getDataLength());
			request.sendRequest(dos);
		}
		catch(IOException e){
			throw new RemoteSPError(e);
		}
	}
}
