/*
 * Java
 *
 * Copyright 2011-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.sp.remote.request;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import ej.sp.ShieldedPlug;

public class ReqWaitFor extends SPInstanceRequest implements SPAsynchronousRequest {

	//Request : <blockID(int)> * blocksIDs.length
	//Response : <blockID(int)> * dataLength/INT_SIZE
	//Throws : IllegalArgumentException, TooManyWaitingThreadsException
	//Return : int[]

	public int[] blocksIDs;
	private Thread requestThread;
	private boolean finished;

	public ReqWaitFor(ShieldedPlug sp, int[] blockIDs) {
		super(sp, REQ_WaitFor);
		this.blocksIDs = blockIDs;
	}

	/**
	 * Returns true if execution and reply of this request must be done when request is received.
	 * Returns false if this request must be executed in another thread.
	 */
	public boolean isSynchronous(){
		return false;
	}

	public int computeDataLengthIntern() {
		return blocksIDs.length * INT_SIZE;
	}

	public void sendRequestIntern(DataOutputStream dos) throws IOException {
		sendBlockIDs(dos);
	}

	protected Object computeReplyResult(DataInputStream dis, int dataLength) throws IOException {
		return readBlockIDS(dis, dataLength);
	}

	public static SPInstanceRequest buildRequestFromStream(ShieldedPlug sp, DataInputStream dis, int dataLength) throws IOException{
		if(SANITY_CHECK)
		{
			spAssert((dataLength&0x3) == 0);//dataLength is multiple of INT_SIZE
		}
		return new ReqWaitFor(sp, readBlockIDS(dis, dataLength));
	}

	public void execute() throws InterruptedException {
		try{
			this.blocksIDs = sp.waitFor(blocksIDs);
		}
		finally {
			synchronized (this) {
				this.finished = true;
				this.notifyAll();
			}
		}
	}

	public int getReplyDataLength() {
		return blocksIDs.length * INT_SIZE;
	}

	public void sendReply(DataOutputStream dos) throws IOException {
		sendBlockIDs(dos);
	}

	public static int[] readBlockIDS(DataInputStream dis, int dataLength) throws IOException {
		int nbBlocks = dataLength / INT_SIZE;
		int[] blocksIDs = new int[nbBlocks];
		for(int i=0 ; i<nbBlocks ; i++) {
			blocksIDs[i] = dis.readInt();
		}
		return blocksIDs;
	}

	public void sendBlockIDs(DataOutputStream dos) throws IOException {
		int[] blockIDs = this.blocksIDs;
		int nbBlocks = blockIDs.length;
		for(int i=0 ; i<nbBlocks ; i++) {
			dos.writeInt(blocksIDs[i]);
		}
	}

	public void waitForEndOfExecution() {
		boolean interrupted = false;

		synchronized (this) {
			while(!this.finished){
				try {
					this.wait();
				} catch (InterruptedException e) {
					// Store the fact that someone ask interruption of the current thread.
					interrupted = true;
				}
			}
		}

		if(interrupted){
			Thread.currentThread().interrupt();
		}
	}

	public void setRequestThread(Thread requestThread) {
		this.requestThread = requestThread;
	}

	public Thread getRequestThread() {
		return requestThread;
	}

}
