/*
 * Java
 *
 * Copyright 2011-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.sp.remote.request;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import com.is2t.sp.remote.RemoteSPError;
import com.is2t.sp.remote.SPProtocol;

import ej.sp.EmptyBlockException;
import ej.sp.TooManyWaitingThreadsException;

abstract public class SPRequest implements SPProtocol{

	public int commandID;
	public int dataLength = -1;//lazy computing

	public boolean replyReceived;
	public Object result;
	public int exceptionId;//!=0 when exception occurred


	public SPRequest(int commandID){
		this.commandID = commandID;
	}

	public int getDataLength(){
		if(dataLength == -1) {
			dataLength = computeDataLength();
		}
		return dataLength;
	}

	public void replyReceived(DataInputStream dis, int dataLength) throws IOException{
		synchronized (this) {
			result = this.computeReplyResult(dis, dataLength);
			replyReceived = true;
			this.notify();
		}
	}

	public void exceptionReceived(int exceptionId){
		synchronized (this) {
			this.exceptionId = exceptionId;
			replyReceived = true;
			this.notify();
		}
	}

	public Object getReplyResult() throws EmptyBlockException{
		if(exceptionId != 0){
			switch(exceptionId){
				case EXC_ILLEGAL_ARGUMENT_EXCEPTION:
					throw new IllegalArgumentException();
				case EXC_SECURITY_EXCEPTION:
					throw new SecurityException();
				case EXC_TOO_MANY_WAITING_THREADS_EXCEPTION:
					throw new TooManyWaitingThreadsException();
				case EXC_EMPTY_BLOCK_EXCEPTION:
					throw new EmptyBlockException();
				case EXC_INDEX_OUT_OF_BOUNDS_EXCEPTION:
					throw new IndexOutOfBoundsException();
				case EXC_ARRAY_INDEX_OUT_OF_BOUNDS_EXCEPTION:
					throw new ArrayIndexOutOfBoundsException();
				default:
					if(SANITY_CHECK) {
						spAssert(false);
					}
			}
		}
		return result;
	}

	public boolean hasResponse(){
		return replyReceived;
	}

	/**
	 * Returns true if execution and reply of this request must be done when request is received.
	 * <p>
	 * If false, the {@link SPRequest} must implement the {@link SPAsynchronousRequest} interface.
	 * Returns false if this request must be executed in another thread.
	 */
	public boolean isSynchronous(){
		return true;
	}

	//Client specific methods
	abstract protected int computeDataLength();
	abstract public void sendRequest(DataOutputStream dos) throws IOException;

	//Server specific methods
	abstract protected Object computeReplyResult(DataInputStream dis, int dataLength) throws IOException;
	abstract public void sendReply(DataOutputStream dos) throws IOException;
	abstract public int getReplyDataLength();

	abstract public void execute() throws EmptyBlockException, InterruptedException;

	public static void spAssert(boolean b){
		if(!b) {
			throw new RemoteSPError();
		}
	}
}
