/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketImpl;

import javax.net.ssl.SSLException;
import javax.net.ssl.SSLHandshakeException;
import javax.net.ssl.SSLSocket;

import com.is2t.support.net.PlainSocketImpl;

import ej.lang.Resource;
import ej.lang.ResourceManager;

public class SSLSocketImpl extends SSLSocket implements Resource {

	private Socket self; // underlying socket
	private int sslID = 0; // the SSL id
	private boolean useClientMode = true; // client mode by default
	private AbstractSSLContext context;
	private boolean autoClose = true;
	private boolean created;
	private boolean initialHandshakeDone = false;
	private boolean handshakingBegun = false;
	private boolean closed = false;
	private boolean needClientAuth;

	private int fd = -1;

	private SSLSocketInputStream input;
	private SSLSocketOuputStream output;
	private final Object handshakeLock = new Object();
	private final Object modeLock = new Object();
	private final Object closeLock = new Object();
	
	private String host = null;

	SSLSocketImpl(AbstractSSLContext context, Socket s, String host, int port, boolean autoClose) throws IOException {
		this.host = host;
		init(context, s);
		if (!s.isConnected()) {
			throw new SocketException("The underlying socket is not connected");
		}
		this.autoClose = autoClose;
		doneConnect();
	}

	SSLSocketImpl(AbstractSSLContext context, String host, int port) throws IOException {
		this.host = host;
		init(context, this);
		SocketAddress socketAddress = host != null ? new InetSocketAddress(host, port)
				: new InetSocketAddress(InetAddress.getByName(null), port);
		connect(socketAddress, 0);
	}

	SSLSocketImpl(AbstractSSLContext context, String host, int port, InetAddress localHost, int localPort)
			throws IOException {
		this.host = host;
		init(context, this);
		bind(new InetSocketAddress(localHost, localPort));
		SocketAddress socketAddress = host != null ? new InetSocketAddress(host, port)
				: new InetSocketAddress(InetAddress.getByName(null), port);
		connect(socketAddress, 0);
	}

	SSLSocketImpl(AbstractSSLContext context) throws IOException {
		init(context, this);
	}

	SSLSocketImpl(AbstractSSLContext context, InetAddress host, int port) throws IOException {
		init(context, this);
		SocketAddress socketAddress = new InetSocketAddress(host, port);
		connect(socketAddress, 0);
	}

	SSLSocketImpl(AbstractSSLContext context, InetAddress address, int port, InetAddress localAddress, int localPort)
			throws IOException {
		init(context, this);
		bind(new InetSocketAddress(localAddress, localPort));
		SocketAddress socketAddress = new InetSocketAddress(address, port);
		connect(socketAddress, 0);
	}

	/*
	 * Package-private constructor used ONLY by SSLServerSocket. The java.net
	 * package accepts the TCP connection after this call is made. This just
	 * initializes handshake state to use "server mode", giving control over the
	 * use of SSL client authentication.
	 */
	SSLSocketImpl(AbstractSSLContext context, boolean serverMode, boolean needClientAuth) throws IOException {
		super((SocketImpl) null);
		init(context, this);
		this.useClientMode = !serverMode;
		this.needClientAuth = needClientAuth;
	}

	private void init(AbstractSSLContext context, Socket self) throws IOException {
		if (context == null) {
			throw new SSLException("No SSL authentification context");
		}
		this.context = context;

		this.self = self;

	}

	@Override
	public void startHandshake() throws IOException {
		doHandShake(true);
	}

	/**
	 * Performs the SSL handshake.
	 *
	 * If the initial handshake is already done and <code>renegotiate</code> is
	 * true, the handshake renegotiation will be performed.
	 *
	 */
	void doHandShake(boolean renegociate) throws IOException {
		synchronized (this.handshakeLock) {
			// check if the socket is connected
			checkConnection();

			if (!this.initialHandshakeDone) {
				// initial handshake
				this.handshakingBegun = true;
				synchronized (this.modeLock) {
					if (this.useClientMode) {
						// client handshake
						SSLSocketSupport.initialClientHandShake(this);
					} else {
						// server handshake
						SSLSocketSupport.initialServerHandShake(this);
					}
				}
				this.initialHandshakeDone = true;
			} else if (renegociate) {
				throw new SSLHandshakeException("handshake renegotiation is not supported");
			}
		}
	}

	int getSSLID() throws SSLException {

		synchronized (closeLock) {
			if (closed) {
				throw new SSLException("socket is closed");
			}
			if (!created) {
				boolean isServer = !this.useClientMode;
				synchronized (this.context) {
					context.beforeCreateSocketCallback(isServer);
					// retrieve the context id according to the mode (client or
					// server)
					int contextID = context.getClientContextID(isServer);
					try {
						sslID = SSLSocketSupport.create(contextID, fd, host, autoClose, !isServer, this.needClientAuth);
						created = true;
						context.afterCreateSocketCallback(isServer, false);
						ResourceManager rm = ResourceManager.getResourceManager();
						if (rm != null) {
							rm.resourceCreated(this);
						}
					} catch (SSLException e) {
						context.afterCreateSocketCallback(isServer, true);
						throw e;
					}
				}
			}
		}
		return sslID;
	}

	boolean isCreated() {
		return created;
	}

	int getFD() {
		return fd;
	}

	void doneConnect() throws IOException {
		// retrieve the socket file descriptor
		fd = fd();
	}

	private int fd() throws IOException {
		if (self == this || !(self instanceof SSLSocketImpl)) {
			SocketImpl impl = self.getImpl();
			if (!(impl instanceof PlainSocketImpl)) {
				throw new IOException("incompatible SocketImpl: " + impl.getClass().getName());
			}
			return ((PlainSocketImpl) impl).getNativeFD();

		} else {
			return ((SSLSocketImpl) self).fd();
		}
	}

	boolean isHandshakeDone() {
		return initialHandshakeDone;
	}

	private void checkConnection() throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		if (!isConnected()) {
			throw new SocketException("Socket is not connected");
		}
	}

	/* OVERRIDDEN Socket methods */

	@Override
	public void bind(SocketAddress bindpoint) throws IOException {
		if (self != this) {
			// if binding on a underlying socket
			throw new IOException("Underlying socket should already be connected");
		}
		super.bind(bindpoint);
	}

	@Override
	public void connect(SocketAddress endpoint) throws IOException {
		connect(endpoint, 0);
	}

	@Override
	public void connect(SocketAddress endpoint, int timeout) throws IOException {
		if (self != this) {
			// if connect on a underlying socket
			throw new IOException("Socket is already connected");
		}
		super.connect(endpoint, timeout);
		doneConnect();
	}

	@Override
	public InetAddress getInetAddress() {
		if (self != this) {
			return self.getInetAddress();
		} else {
			return super.getInetAddress();
		}
	}

	@Override
	public InetAddress getLocalAddress() {
		if (self != this) {
			return self.getLocalAddress();
		} else {
			return super.getLocalAddress();
		}
	}

	@Override
	public int getPort() {
		if (self != this) {
			return self.getPort();
		} else {
			return super.getPort();
		}
	}

	@Override
	public int getLocalPort() {
		if (self != this) {
			return self.getLocalPort();
		} else {
			return super.getLocalPort();
		}
	}

	@Override
	public SocketAddress getRemoteSocketAddress() {
		if (self != this) {
			return self.getRemoteSocketAddress();
		} else {
			return super.getRemoteSocketAddress();
		}
	}

	@Override
	public SocketAddress getLocalSocketAddress() {
		if (self != this) {
			return self.getLocalSocketAddress();
		} else {
			return super.getLocalSocketAddress();
		}
	}

	@Override
	public InputStream getInputStream() throws IOException {
		checkConnection();
		synchronized (this) {
			if (input == null) {
				input = new SSLSocketInputStream(this);
			}
		}
		return input;
	}

	@Override
	public OutputStream getOutputStream() throws IOException {
		checkConnection();
		synchronized (this) {
			if (output == null) {
				output = new SSLSocketOuputStream(this);
			}
		}
		return output;
	}

	@Override
	public void setTcpNoDelay(boolean on) throws SocketException {
		if (self != this) {
			self.setTcpNoDelay(on);
		} else {
			super.setTcpNoDelay(on);
		}
	}

	@Override
	public boolean getTcpNoDelay() throws SocketException {
		if (self != this) {
			return self.getTcpNoDelay();
		} else {
			return super.getTcpNoDelay();
		}
	}

	@Override
	public synchronized void setSoTimeout(int timeout) throws SocketException {
		if (self != this) {
			self.setSoTimeout(timeout);
		} else {
			super.setSoTimeout(timeout);
		}
	}

	@Override
	public synchronized int getSoTimeout() throws SocketException {
		if (self != this) {
			return self.getSoTimeout();
		} else {
			return super.getSoTimeout();
		}
	}

	@Override
	public synchronized void setSendBufferSize(int size) throws SocketException {
		if (self != this) {
			self.setSendBufferSize(size);
		} else {
			super.setSendBufferSize(size);
		}
	}

	@Override
	public synchronized int getSendBufferSize() throws SocketException {
		if (self != this) {
			return self.getSendBufferSize();
		} else {
			return super.getSendBufferSize();
		}
	}

	@Override
	public synchronized void setReceiveBufferSize(int size) throws SocketException {
		if (self != this) {
			self.setReceiveBufferSize(size);
		} else {
			super.setReceiveBufferSize(size);
		}
	}

	@Override
	public synchronized int getReceiveBufferSize() throws SocketException {
		if (self != this) {
			return self.getReceiveBufferSize();
		} else {
			return super.getReceiveBufferSize();
		}
	}

	@Override
	public void setKeepAlive(boolean on) throws SocketException {
		if (self != this) {
			self.setKeepAlive(on);
		} else {
			super.setKeepAlive(on);
		}
	}

	@Override
	public boolean getKeepAlive() throws SocketException {
		if (self != this) {
			return self.getKeepAlive();
		} else {
			return super.getKeepAlive();
		}
	}

	@Override
	public void setTrafficClass(int tc) throws SocketException {
		if (self != this) {
			self.setTrafficClass(tc);
		} else {
			super.setTrafficClass(tc);
		}
	}

	@Override
	public int getTrafficClass() throws SocketException {
		if (self != this) {
			return self.getTrafficClass();
		} else {
			return super.getTrafficClass();
		}
	}

	@Override
	public synchronized void close() throws IOException {
		synchronized (closeLock) {

			if (closed) {
				// already closed
				return;
			}

			try {
				if (created) {
					SSLSocketSupport.close(this, autoClose);
				}
				if (self == this) {
					super.close();
				} else {
					if (autoClose) {
						self.close();
					}
				}
			} catch (SSLException e) {
				// an exception occurred in the SSL close operation, fd has to
				// be closed
				if (self == this) {
					super.close();
				} else {
					if (autoClose) {
						self.close();
					}
				}
			} finally {
				if (created) {
					try {
						SSLSocketSupport.freeSSL(this);
					} catch (Exception e2) {
						// ignore
					}
					context.closeSocketCallback(!useClientMode);
				}
				ResourceManager rm = ResourceManager.getResourceManager();
				if (rm != null) {
					rm.resourceReclaimed(this);
				}
				sslID = -1;
				closed = true;
			}
		}
	}

	@Override
	public boolean isConnected() {
		if (self != this) {
			return self.isConnected();
		} else {
			return super.isConnected();
		}
	}

	@Override
	public boolean isBound() {
		if (self != this) {
			return self.isBound();
		} else {
			return super.isBound();
		}
	}

	@Override
	public boolean isClosed() {
		synchronized (closeLock) {
			return closed;
		}
	}

	@Override
	public Object getSource() {
		return this;
	}

	@Override
	public void reclaim() {
		try {
			close();
		} catch (IOException e) {
		}
	}

	/**
	 * Sets the flag controlling whether a server mode socket *REQUIRES* SSL
	 * client authentication.
	 *
	 * As long as handshaking has not started, we can change whether client
	 * authentication is needed. Otherwise, we will need to wait for the next
	 * handshake.
	 */
	@Override
	public void setNeedClientAuth(boolean need) {
		if (!useClientMode && !handshakingBegun) {
			this.needClientAuth = need;
		}

	}

	@Override
	public boolean getNeedClientAuth() {
		return needClientAuth;
	}

	@Override
	public void setUseClientMode(boolean mode) {
		synchronized (modeLock) {
			if (handshakingBegun) {
				throw new IllegalArgumentException("handshaking has begun");
			}
			useClientMode = mode;
		}
	}

	@Override
	public boolean getUseClientMode() {
		return useClientMode;
	}
}
