/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2016-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl;

import java.io.IOException;
import java.net.SocketTimeoutException;

import javax.net.ssl.SSLException;
import javax.net.ssl.SSLHandshakeException;

import com.is2t.support.net.NetworkResources;
import com.is2t.support.net.ssl.natives.SSLSocketSupportNatives;

import ej.bon.Immortals;
import ej.error.Message;
import ej.sni.SNI;

public class SSLSocketSupport {

	private static boolean initialized = false;

	static {
		initialize();
	}

	public static void initialize() {
		if (!initialized) {
			int result = SSLSocketSupportNatives.initialize();
			if (result < 0) {
				throw new RuntimeException("SSL initialization failed");
			}
		}
		initialized = true;
	}

	public static int create(int contextid, int fd, String host, boolean autoclose, boolean isClientMode, boolean needClientAuth)
			throws SSLException {

		synchronized (SSLResources.HOSTNAME_BUFFER) {
			byte[] hostname = SSLResources.HOSTNAME_BUFFER;
			int hostLength = 0;
			
			if (host != null) {
				try {
					SNI.toCString(host, hostname);
					hostLength = host.length() + 1; //+1 for null-terminated
				} catch (IndexOutOfBoundsException e) {
					throw new SSLException("Could not set the hostname");
				}
			}
			int ret = SSLSocketSupportNatives.create(contextid, fd, hostname, hostLength, autoclose, isClientMode, needClientAuth, false);
			
			while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				ret = SSLSocketSupportNatives.create(contextid, fd, hostname, hostLength, autoclose, isClientMode, needClientAuth, true);
			}
			if (ret == SSLErrorConstants.J_CREATE_SSL_ERROR) {
				throw new SSLException("Could not create the underlying SSL Socket");
			}
			
			return ret;
		}
	}

	public static void close(SSLSocketImpl socket, boolean autoclose) throws SSLException {
		int ret = -1;
		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		ret = SSLSocketSupportNatives.close(sslid, fd, autoclose, false);
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLSocketSupportNatives.close(sslid, fd, autoclose, true);
		}

		if (ret < 0) {
			throw new SSLException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static void freeSSL(SSLSocketImpl socket) throws SSLException {
		int ret = -1;
		int sslid = socket.getSSLID();
		ret = SSLSocketSupportNatives.freeSSL(sslid, false);
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLSocketSupportNatives.freeSSL(sslid, true);
		}

		if (ret < 0) {
			throw new SSLException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static void initialClientHandShake(SSLSocketImpl socket) throws SSLException {

		int ret = -1;
		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		ret = SSLSocketSupportNatives.initialClientHandShake(sslid, fd, false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {

			ret = SSLSocketSupportNatives.initialClientHandShake(sslid, fd, true);
		}

		if (ret < 0) {
			throw new SSLHandshakeException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static void initialServerHandShake(SSLSocketImpl socket) throws SSLException {
		int ret = -1;
		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		ret = SSLSocketSupportNatives.initialServerHandShake(sslid, fd, false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {

			ret = SSLSocketSupportNatives.initialServerHandShake(sslid, fd, true);
		}

		if (ret < 0) {
			throw new SSLHandshakeException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static int read(SSLSocketImpl socket, byte[] buf, int offset, int length) throws IOException {
		int ret;
		// verify data array type to optimize native call if array is an
		// immortal one
		boolean isImmortalData = Immortals.isImmortal(buf);
		byte[] immortal;
		int len;
		int immortalOffset;
		if (isImmortalData) {
			immortal = buf;
			len = length;
			immortalOffset = offset;
		} else {
			immortal = NetworkResources.reserveIOResource();
			len = Math.min(length, immortal.length);
			immortalOffset = 0;
		}
		try {

			int sslid = socket.getSSLID();
			int fd = socket.getFD();
			ret = SSLSocketSupportNatives.read(sslid, fd, immortal, immortalOffset, len, false);

			while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {

				ret = SSLSocketSupportNatives.read(sslid, fd, immortal, immortalOffset, len, true);
			}

			if (ret > 0) {
				// optimization: no need to make copy if data is an immortal
				// array
				if (!isImmortalData) {
					System.arraycopy(immortal, immortalOffset, buf, offset, ret);
				}
			}

			if (ret < SSLErrorConstants.J_EOF) {

				if (ret == SSLErrorConstants.J_SOCKET_TIMEOUT) {
					// timeout exception
					throw new SocketTimeoutException();
				}

				// error
				throw new IOException(Message.at(SSLErrorMessages.getInstance(), ret));
			}

			return ret;
		} finally {
			// do not release write buffer immortal array if not allocated in
			// the resources pool
			if (!isImmortalData) {
				NetworkResources.releaseIOResource(immortal);
			}
		}
	}

	public static int available(SSLSocketImpl socket) throws IOException {
		int ret = SSLSocketSupportNatives.available(socket.getSSLID(), false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLSocketSupportNatives.available(socket.getSSLID(), true);
		}

		if (ret < 0) {
			throw new IOException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
		return ret;
	}

	public static int write(SSLSocketImpl socket, byte[] buf, int offset, int length) throws IOException {

		int writtenBytes = 0;

		int immortalOffset = 0;
		// verify data array type to optimize native call if array is an
		// immortal one
		boolean isImmortalData = Immortals.isImmortal(buf);
		byte[] immortal;
		if (isImmortalData) {
			immortal = buf;
		} else {
			immortal = NetworkResources.reserveIOResource();
		}
		try {
			int sslid = socket.getSSLID();
			int fd = socket.getFD();
			int remaining;
			// while there is data to write, transmit by "IMMORTAL capacity" (at
			// most) chunks
			while ((remaining = (length - writtenBytes)) > 0) { // we did not
				// reach the end
				// of buffer
				// put into IMMORTAL
				int len = Math.min(remaining, immortal.length);
				// optimization: no need to make copy if data is an immortal
				// array
				if (!isImmortalData) {
					System.arraycopy(buf, offset, immortal, immortalOffset, len);
				} else {
					immortalOffset = offset;
				}

				System.arraycopy(buf, offset, immortal, 0, len);

				int ret = SSLSocketSupportNatives.write(sslid, fd, immortal, immortalOffset, len, false);

				while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
					ret = SSLSocketSupportNatives.write(sslid, fd, immortal, immortalOffset, len, true);
				}
				if (ret <= 0) {
					throw new IOException(Message.at(SSLErrorMessages.getInstance(), ret));
				}
				writtenBytes += ret;
				offset += ret;
			}
			return writtenBytes;
		} finally {
			// do not release write buffer immortal array if not allocated in
			// the resources pool
			if (!isImmortalData) {
				NetworkResources.releaseIOResource(immortal);
			}
		}

	}

}
