/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl.natives;

import com.is2t.support.net.ssl.SSLErrorConstants;

public class SSLContextSupportNatives {

	/**
	 * Creates a new client SSL Context based on the given SSL/TLS protocol
	 * version.
	 *
	 * @param protocol
	 *            the SSL/TLS protocol version (0 for SSLv3, 1 for TLSv1, 2 for
	 *            TLSv1.1 and 3 for TLSv1.2).
	 * @param isClientContext
	 *            controls whether the underlying context will be created to
	 *            manage ssl client or ssl server sockets. <code>true</code> for
	 *            client sockets; <code>false</code> for server sockets.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return the new SSL Context ID on success,
	 *         {@link SSLErrorConstants#J_CREATE_SSL_CONTEXT_ERROR} on error
	 *         (for example if not enough space to create the context).
	 */
	public static native int createContext(int protocol, boolean isClientContext, boolean retry);

	/**
	 * Releases the SSL Context identified by the given contextID.
	 *
	 * @param contextID
	 *            the SSL context ID.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 *
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code.
	 */
	public static native int closeContext(int contextID, boolean retry);

	/**
	 * Adds the given certificate as trusted certificate into the given SSL
	 * context.
	 *
	 * @param contextID
	 *            the SSL Context ID.
	 * @param cert
	 *            the certificate buffer.
	 * @param off
	 *            the offset in the buffer at which the certificate content
	 *            started.
	 * @param len
	 *            the certificate content length.
	 * @param format
	 *            the certificate encoded format type (0 for PEM format and 1
	 *            for DER format).
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code.
	 */
	public static native int addTrustedCert(int contextID, byte[] cert, int off, int len, int format, boolean retry);

	/**
	 * Removes all trusted certificates from the given SSL context
	 *
	 * @param contextID
	 *            the SSL Context ID
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code
	 */
	public static native int clearTrustStore(int contextID, boolean retry);

	/**
	 * Removes all credentials (private key and associated certificates) from
	 * the given ssl context
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code.
	 */
	public static native int clearKeyStore(int contextID, boolean retry);

	/**
	 * Sets the certificate of the context private key.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param cert
	 *            the certificate of the context private key.
	 * @param offset
	 *            the offset in the buffer at which the certificate content
	 *            started.
	 * @param len
	 *            the certificate content length.
	 * @param format
	 *            the certificate encoded format type (0 for PEM format and 1
	 *            for DER format).
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code.
	 *
	 */
	public static native int setCertificate(int contextID, byte[] cert, int offset, int len, int format, boolean retry);

	/**
	 * Sets the context private key.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param privateKey
	 *            the context private key.
	 * @param offset
	 *            the offset in the buffer at which the private key content
	 *            started.
	 * @param len
	 *            the private key content length.
	 * @param password
	 *            the password to recover the private key if it is encrypted.
	 * @param passwordOffset
	 *            the offset in the password buffer at which the password
	 *            content started.
	 * @param passwordLen
	 *            the password length.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code.
	 */
	public static native int setPrivateKey(int contextID, byte[] privateKey, int offset, int len, byte[] password,
			int passwordOffset, int passwordLen, boolean retry);


	/**
	 * Initializes the chain buffer and returns its total size.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param nbChainCerts
	 *            the number of certificates in the chain.
	 * @param chainCertsTotalSize
	 *            the total size of certificates in the chain.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return the total size of the chain buffer on success or a negative error
	 *         code.
	 */
	public static native int initChainBuffer(int contextID, int nbChainCerts, int chainCertsTotalSize, boolean retry);

	/**
	 * Adds the given certificate to the certificate chain buffer associated
	 * with the context private key.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param cert
	 *            the certificate to be added in the chain.
	 * @param offset
	 *            the offset in the buffer at which the certificate content
	 *            started.
	 * @param len
	 *            the certificate content length.
	 * @param format
	 *            the certificate encoded format type (0 if PEM format and 1 if
	 *            CERT_DER_FORMAT).
	 * @param chainBufferSize
	 *            the chain buffer size.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code.
	 */
	public static native int addChainCertificate(int contextID, byte[] cert, int offset, int len, int format,
			int chainBufferSize, boolean retry);


}
