/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl.natives;

import com.is2t.support.net.ssl.SSLErrorConstants;

public class SSLSocketSupportNatives {

	/**
	 * Initializes SSL/TLS stack components
	 *
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success
	 */
	public static native int initialize();

	/**
	 * Creates a new SSL Socket using the given SSL Context and the given
	 * underlying Socket file descriptor {@code fd} for I/O operations
	 *
	 * @param contextID
	 *            the SSL Context ID
	 * @param fd
	 *            the underlying socket file descriptor used for I/O operations
	 * @param hostname
	 *            the server hostname
	 * @param hostnameLen
	 *            the length of the server hostname
	 * @param autoclose
	 *            true if the underlying socket will be closed when this SSL
	 *            Socket is closed; false otherwise
	 * @param useClientMode
	 *            configures the socket to use client (or server) mode when
	 *            handshaking. <code>true</code> if the socket should start its
	 *            handshaking in "client" mode; <code>false</code> if the socket
	 *            should start its handshaking in "server" mode
	 * @param needClientAuth
	 *            configures the socket to require or not the client
	 *            authentication. This option is only useful for sockets in the
	 *            server mode.<code>true</code> if the client authentication is
	 *            required; <code>false</code> otherwise.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return the new SSL Socket ID on success,
	 *         {@link SSLErrorConstants#J_CREATE_SSL_SOCKET_ERROR} on error
	 */
	public static native int create(int contextID, int fd, byte[] hostname, int hostnameLen,
			boolean autoclose, boolean useClientMode, boolean needClientAuth, boolean retry);

	/**
	 * Sends close_notify alert and reads the received close_notify to clear the
	 * socket input stream if autoclose is set to false.
	 *
	 * @param sslID
	 *            the SSL Socket ID
	 * @param fd
	 *            the associated underlying socket file descriptor
	 * @param autoclose
	 *            if true, no need to clear the input stream, the underlying
	 *            socket will be closed; otherwise read the received
	 *            close_notify alert to clear the socket input stream.
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code
	 */
	public static native int close(int sslID, int fd, boolean autoclose, boolean retry);

	/**
	 * Releases the SSL Socket identified by the given ssl ID.
	 *
	 * @param sslID
	 *            the SSL Socket ID
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code
	 */
	public static native int freeSSL(int sslID, boolean retry);

	/**
	 * Initiates the handshaking in client mode
	 *
	 * @param sslID
	 *            the SSL Socket ID
	 * @param fd
	 *            the associated underlying socket file descriptor
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code
	 */
	public static native int initialClientHandShake(int sslID, int fd, boolean retry);

	/**
	 * Initiates the handshaking in server mode
	 *
	 * @param sslID
	 *            the SSL Socket ID
	 * @param fd
	 *            the associated underlying socket file descriptor
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return {@link SSLErrorConstants#J_SSL_NO_ERROR} on success or a negative
	 *         error code
	 */
	public static native int initialServerHandShake(int sslID, int fd, boolean retry);

	/**
	 * Reads len bytes from the SSL Socket internal read buffer into the
	 * specified buffer {@code buf}
	 *
	 * @param sslID
	 *            the SSL Socket ID
	 * @param fd
	 *            the associated underlying socket file descriptor
	 * @param buf
	 *            the buffer into which the data is read
	 * @param off
	 *            the start offset in the buffer at which the data is written
	 * @param len
	 *            the maximum number of data to read
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return the number of bytes read on success;
	 *         {@link SSLErrorConstants#J_EOF} if there is no more data because
	 *         the end of the stream has been reached or a negative error code
	 */
	public static native int read(int sslID, int fd, byte[] buf, int off, int len, boolean retry);

	/**
	 * Returns the number of bytes available in the SSL Socket internal read
	 * buffer
	 *
	 * @param sslID
	 *            the SSL Socket ID
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return the number of available bytes on success or a negative error code
	 */
	public static native int available(int sslID, boolean retry);

	/**
	 * Writes len bytes from the specified buffer {@code buf} starting at the
	 * offset to the SSL Socket output stream
	 *
	 * @param sslID
	 *            the SSL Socket ID
	 * @param fd
	 *            the associated underlying socket file descriptor
	 * @param buf
	 *            the buffer of data to write
	 * @param off
	 *            the offset in the buffer
	 * @param len
	 *            the number of data to write
	 * @param retry
	 *            true if the calling process repeats the call to this operation
	 *            for its completion when the previous call has returned
	 *            {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that
	 *            the operation was not completed.
	 * @return the number of bytes written on success or a negative error code
	 */
	public static native int write(int sslID, int fd, byte[] buf, int off, int len, boolean retry);

}
