/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl.x509;

import java.security.Key;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.util.Enumeration;

import javax.net.ssl.X509KeyManager;

import com.is2t.support.net.ssl.SSLContextSupport;
import com.is2t.support.security.x509.X509CertImpl;
import com.is2t.support.security.x509.X509KeyImpl;

public class X509KeyManagerImpl implements X509KeyManager {

	private X509KeyImpl privKey;
	private X509CertImpl[] certificates;
	private String password;

	public X509KeyManagerImpl(KeyStore ks, char[] password)
			throws KeyStoreException, UnrecoverableKeyException, NoSuchAlgorithmException {
		privKey = null;
		certificates = null;
		this.password = null;
		// credentialsMap = new HashMap<String, X509Credentials>();
		if (ks == null) {
			return;
		}

		// search the first credentials whose key and certificate are
		// respectively an implementation of X509KeyImpl and X509CertImpl
		for (Enumeration<String> aliases = ks.aliases(); aliases.hasMoreElements();) {
			String alias = aliases.nextElement();
			if (!ks.isKeyEntry(alias)) {
				continue;
			}
			Key key = ks.getKey(alias, password);
			if (!(key instanceof X509KeyImpl)) {
				continue;
			}
			Certificate[] certs = ks.getCertificateChain(alias);
			if ((certs == null) || (certs.length == 0) || !(certs[0] instanceof X509CertImpl)) {
				continue;
			}
			if (!(certs instanceof X509CertImpl[])) {
				Certificate[] tmp = new X509CertImpl[certs.length];
				// TODO be sure all certificates are instanceof X509CertImpl
				// before copy!
				System.arraycopy(certs, 0, tmp, 0, certs.length);
				certs = tmp;
			}
			privKey = (X509KeyImpl) key;
			certificates = (X509CertImpl[]) certs;
			if (password != null) {
				this.password = new String(password);
			}
			return;
		}
	}

	public void initContextKeyStore(int contextID) throws KeyManagementException, CertificateEncodingException {
		// clear the context key store
		SSLContextSupport.clearKeyStore(contextID);
		// init with the private key and the associated chain certificates
		SSLContextSupport.useCredentials(contextID, privKey, password, certificates);
	}
}
