/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl.x509;

import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.util.ArrayList;
import java.util.Enumeration;

import javax.net.ssl.X509TrustManager;

import com.is2t.support.net.ssl.SSLContextSupport;
import com.is2t.support.security.x509.X509CertImpl;


public class X509TrustManagerImpl implements X509TrustManager {

	private final ArrayList<X509CertImpl> trustedCerts;

	public X509TrustManagerImpl(KeyStore ks) {

		trustedCerts = getTrustedCerts(ks);

	}

	private ArrayList<X509CertImpl> getTrustedCerts(KeyStore ks) {
		ArrayList<X509CertImpl> trustedCerts = new ArrayList<X509CertImpl>();

		if (ks != null) {
			try {
				for (Enumeration<String> e = ks.aliases(); e.hasMoreElements();) {
					String alias = e.nextElement();
					if (ks.isCertificateEntry(alias)) {
						Certificate cert = ks.getCertificate(alias);
						if (cert instanceof X509CertImpl) {
							// get only X509CertImpl certificate implementation
							trustedCerts.add((X509CertImpl) cert);
						}
					} else if (ks.isKeyEntry(alias)) {
						Certificate[] certs = ks.getCertificateChain(alias);
						if ((certs != null) && (certs.length > 0) && (certs[0] instanceof X509CertImpl)) {
							trustedCerts.add((X509CertImpl) certs[0]);
						}
					}
				}
			} catch (KeyStoreException e) {
				// ignore
			}
		}
		return trustedCerts;
	}

	public void initContextTrustStore(int contextID) throws KeyManagementException, CertificateEncodingException {
		// clear the context trust store
		SSLContextSupport.clearTrustStore(contextID);
		int nbCerts = trustedCerts.size();
		for (int i = -1; ++i < nbCerts;) {
			SSLContextSupport.addTrustedCert(contextID, trustedCerts.get(i));
		}
	}
}
