/*
 * Java
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.support.security;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;

public abstract class NativeResource extends WeakReference {

	/** Linked list of opened resources. */
	private static NativeResource OpenedResourcesList;
	/** weak reference queue used to collect and close useless native resources */
	private static final ReferenceQueue<NativeResource> Queue = new ReferenceQueue();

	/** the native identifier related to this resource **/
	private int nativeId;

	/** previous weak reference **/
	private NativeResource prev;
	/** next weak reference **/
	private NativeResource next;

	public NativeResource(Object ref, int nativeId) {
		super(ref, Queue);
		this.nativeId = nativeId;
		add();
	}

	public final synchronized void close() {
		int nativeId = this.nativeId;
		if (nativeId != -1) {
			remove();
			closeNativeResource(nativeId);
			this.nativeId = -1;
		}
	}

	abstract protected void closeNativeResource(int nativeId);

	/**
	 * Adds this weak reference to the opened resource list.
	 */
	private void add() {
		synchronized (Queue) {
			this.prev = null;
			this.next = OpenedResourcesList;
			if (OpenedResourcesList != null) {
				OpenedResourcesList.prev = this;
			}
			OpenedResourcesList = this;
		}
	}

	/**
	 * Removes this weak reference from the opened resource list.
	 */
	private void remove() {
		synchronized (Queue) {
			NativeResource prev = this.prev;
			NativeResource next = this.next;
			if (prev == null) {
				// the current reference is first element in the list
				OpenedResourcesList = next;
			} else {
				prev.next = next;
			}
			if (next != null) {
				next.prev = prev;
			}
			this.prev = null;
			this.next = null;
		}
	}

	/**
	 * Releases all useless native resources. A resource becomes useless when it is
	 * garbage collected.
	 *
	 * @param triggerGC
	 *            flag indicating whether the GC must be triggered to detect new
	 *            potential useless native resources.
	 */
	public static void recycle(boolean triggerGC) {
		if (triggerGC) {
			// garbage useless resources
			System.gc();
		}

		synchronized (Queue) {
			NativeResource resourceToClose = (NativeResource) Queue.poll();
			while (resourceToClose != null) {
				resourceToClose.close();
				resourceToClose = (NativeResource) Queue.poll();
			}
		}
	}

}
