/*
 * Java
 *
 * Copyright 2010-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.microui.display;

import com.is2t.microbsp.microui.natives.NSystemPump;
import com.is2t.microui.watchdog.KFWatchdog;

import ej.bon.Constants;
import ej.kf.DeadFeatureException;
import ej.kf.Feature;
import ej.kf.Kernel;
import ej.kf.Module;
import ej.microui.MicroUIProperties;
import ej.microui.SystemMicroUI;

public class KFDisplayPump extends DisplayPump {

	/**
	 * Event to switch display when starting/stopping a feature.
	 */
	/* default */ static final int EVENT_DISPLAY_SWITCH_DISPLAY = EVENT_DISPLAY | 0x07;

	/* default */ Display display;

	/**
	 * Watchdog started in KF display pump when pumping. If pump event execution is
	 * taking too much time, it kills the feature that generate the event.
	 */
	private final KFWatchdog pumpWatchdog = new KFWatchdog();

	public KFDisplayPump(Display display) {
		super();
		this.display = display;
	}

	@Override
	protected Display getDisplay() {
		return this.display;
	}

	@Override
	public boolean handleEvent(int event) {
		// Adding a new event is done in Kernel mode.
		Kernel.enter();
		return super.handleEvent(event);
	}

	@Override
	public boolean isCurrentDisplay(Display d) {
		Kernel.enter();
		return this.display == d;
	}

	@Override
	protected void executeEvent(int event, int eventType) throws InterruptedException {
		this.pumpWatchdog.start(Constants.getInt(MicroUIProperties.PUMP_WATCHDOG_MAX_DELAY),
				Kernel.getOwner(this.display));

		if (eventType == EVENT_DISPLAY_SWITCH_DISPLAY) {
			((Runnable) (getEventData(event))).run();
		} else {
			super.executeEvent(event, eventType);
		}
	}

	@Override
	protected void executeDone() {
		this.pumpWatchdog.close();
		super.executeDone();
	}

	/**
	 * Switch the DisplayPump owner. This action is serialized by DisplayPump
	 * thread. The display pump queue is cleared. This function waits until the
	 * current event is terminated, so that when this method returns, the queue is
	 * empty, the last event has been fully executed and the current display has
	 * been set.
	 */
	@Override
	public void switchDisplay(Display newDisplay) {
		// here, permission is granted => add event to set the current display
		Kernel.enter();

		// true if the display switch must be executed now in the current thread, false
		// otherwise.

		KFSetCurrentDisplayRunnable setCurrentDisplayEvent;
		boolean isDPThread = Thread.currentThread() == getThread();
		boolean executeNow = isDPThread;

		synchronized (this) { // sync on runnable buffer

			// create new KFSetCurrentDisplayRunnable event
			setCurrentDisplayEvent = new KFSetCurrentDisplayRunnable(this, newDisplay);
			int event = SystemMicroUI.buildEvent(EVENT_DISPLAY_SWITCH_DISPLAY, storeEventData(setCurrentDisplayEvent));

			// add new event or replace old one
			int oldFirstEventInQueue = NSystemPump.manageKernelEvent(event,
					SystemMicroUI.getTypeMask(EVENT_DISPLAY_SWITCH_DISPLAY),
					!executeNow /* have to reset queue and add event if new event cannot be executed now */);

			if (oldFirstEventInQueue != 0) {
				// there was a KFSetCurrentDisplayRunnable event at index 0; this event has been
				// replaced by new one; remove associated display in objects list

				// this call gets object and removes the object from list
				getEventData(oldFirstEventInQueue);
			}
			// else: there was no KFSetCurrentDisplayRunnable event in queue
		}

		if (executeNow) {
			// Execute the display switch now.
			// We are no more synchronized on a Kernel object.
			setCurrentDisplayEvent.run();
		} else {
			// Here, the thread is not the DiplayPump
			// Wait for execution of CLEAR event
			setCurrentDisplayEvent.waitForExecution();
		}
	}

	private void requestPhysicalDisplayAccess(Module module) {
		final Display display = this.display;
		Module displayOwner = Kernel.getOwner(display);
		if (displayOwner != module) {
			// event owned by a Feature: only added if it is the current owner of the Pump
			// => need to request the Display access

			Kernel.runUnderContext(module, new Runnable() {
				@Override
				public void run() {
					Display.getDisplay().requestPhysicalDisplayAccess();
				}
			});
		}
	}

	@Override
	public void createAndHandleEvent(int event) {
		// Adding a new event is done in Kernel mode.
		// Caller has not called Kernel.enter() for sure
		Module m = Kernel.getContextOwner();
		Kernel.enter();
		requestPhysicalDisplayAccess(m);
		super.createAndHandleEvent(event);
	}

	@Override
	public void createAndHandleEvent(int event, Object eventData) {
		// Adding a new event is done in Kernel mode.
		Kernel.enter();
		requestPhysicalDisplayAccess(Kernel.getOwner(eventData));
		super.createAndHandleEvent(event, eventData);
	}

	@Override
	protected void handleError(Throwable e) {

		/**
		 * In case pump watchdog has been triggered, pump will catch a DFE. In that
		 * case, since KF has no dead listener yet, we ignore the exception in order to
		 * prevent to add the following code in kernel:
		 *
		 * <code>
		 * Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
		 * 	public void uncaughtException(Thread t, Throwable e) {
		 * 		if (!MicroUI.isUIThread(t) || e.getClass() != DeadFeatureException.class) {
		 * 			e.printStackTrace();
		 * 		}
		 * 		// else: Feature's MicroUI pump correctly released.
		 * 	}
		 * });
		 * </code>
		 *
		 */

		try {
			if (!(e instanceof DeadFeatureException)) {
				super.handleError(e);
			}
		} catch (Throwable ex) {
			// An application error occurred but error handling failed:
			// Possible reasons:
			// - the owner of the error (= owner of the Display) has been killed
			// - the Display has been detached (NPE when executing events in the
			// queue)
			// => silently skip
		}
	}

	/**
	 * Detach the Display if it is owned by the given Feature and attach it to
	 * default Kernel one
	 */
	public synchronized void resetDisplayIfOwnedBy(Feature feature, Display kernelDisplay) {
		if (Kernel.getOwner(this.display) == feature) {
			setCurrentDisplay(kernelDisplay);
		}
	}

	/**
	 * Change the {@link Display} currently connected to the Pump and clear the
	 * internal state.
	 */
	public synchronized void setCurrentDisplay(Display d) {
		clearQueue();
		d.resetFlush();
		this.display = d;
	}

	@Override
	void callOnFlushCompleted(Runnable r) {
		if (Kernel.getOwner(this.display) == (null == r ? Kernel.getContextOwner() : Kernel.getOwner(r))) {
			// Adding a new event is done in Kernel mode.
			Kernel.enter();
			super.callOnFlushCompleted(r);
		}
		// else: caller cannot add its runnable or remove the current runnable
	}
}
