/*
 * Java
 *
 * Copyright 2017-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.microui.display;

import ej.annotation.Nullable;
import ej.bon.Constants;
import ej.microui.MicroUIProperties;

/**
 * This class associates a string with a font. The string rendering is
 * accelerated because the couple string/font is fixed.
 */
public class RenderableString {

	/**
	 * This default index is used when the string indexes have not been computed
	 * yet.
	 */
	private static final int INVALID_INDEX = -1;

	/* default */ final String string;
	/* default */ final int[] indices;
	/* default */ final Font font;
	/* default */ final int width;
	/* default */ @Nullable
	char[] chars; // null init; useless until first call to getCharArray()

	/**
	 * microUI-API
	 *
	 * @param string
	 *            microUI-API
	 * @param font
	 *            microUI-API
	 */
	public RenderableString(String string, Font font) {
		assert string != null && font != null;

		this.string = string;
		this.font = font;

		// create array of indexes; initializes it with invalid indexes in order
		// to force the native side to retrieve all characters indexes
		int l = string.length(); // NPE
		this.indices = new int[l];
		for (int i = l; --i >= 0;) {
			this.indices[i] = INVALID_INDEX;
		}

		// get width without ratio (this size will never change)
		char[] data;
		int offset;
		int length;
		if (Constants.getBoolean(MicroUIProperties.EDC_INTERNAL)) {
			data = this.string.chars;
			offset = this.string.offset;
			length = this.string.length;
		} else {
			data = Font.getStringChars(string);
			offset = 0;
			length = data.length;
		}
		this.width = font.getWidth(data, offset, length, this.indices); // NPE
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public String getString() {
		return this.string;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public char[] getCharArray() {
		// useless to create this array systematically

		char[] array = this.chars;

		// first check prevents to sync at each call
		if (array == null) {

			synchronized (this) {

				// second check is required when two threads are trying to get array at same
				// time (for the first time only)
				array = this.chars;
				if (array == null) {
					this.chars = this.string.toCharArray();
					array = this.chars;
				}
			}
		}

		return array;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int[] getIndices() {
		return this.indices;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public Font getFont() {
		return this.font;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getWidth() {
		return this.width;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getHeight() {
		return this.font.getHeight();
	}
}
