/**
 * Java
 *
 * Copyright 2011-2015 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.sp;

import com.is2t.natives.sp.NShieldedPlug;

import ej.bon.ByteArray;
import ej.bon.IllegalStateException;

public class ShieldedPlug {

	private static final int INT_SIZE = 4;
	private static final int LONG_SIZE = 8;
	private static final int FLOAT_SIZE = 4;
	private static final int DOUBLE_SIZE = 8;

	// class
	public static SimpleIntToObjectTable JavaDatabases;// = new SimpleIntToObjectTable(); // create it here ?

	// instance
	public int databasePtr;
	public SimpleIntToObjectTable blocks = new SimpleIntToObjectTable(); // create it here ?

	/**
	 * Could not create a <code>ShieldedPlug</code>
	 */
	private ShieldedPlug(int ID, int databasePtr) {
		this.databasePtr = databasePtr;
	}

	/**
	 * Returns the database defined at the given ID.
	 * 
	 * @param ID
	 *            identification number of the requested database
	 * @return the database with the given ID
	 * @throws IllegalArgumentException
	 *             if no database is defined with the given ID
	 */
	public synchronized static ShieldedPlug getDatabase(int ID) {
		if (JavaDatabases == null) {
			JavaDatabases = new SimpleIntToObjectTable();
		}

		Object database = JavaDatabases.at(ID);
		if (database == null) {
			// not found, get from native
			int databasePtr = NShieldedPlug.getDatabasePtr(ID);
			if (databasePtr == 0) {
				// no database found with this ID
				throw new IllegalArgumentException();
			}
			database = new ShieldedPlug(ID, databasePtr);
			JavaDatabases.atPutAlreadyCheckNotExists(ID, database);
		}
		return (ShieldedPlug) database;
	}

	/**
	 * Creates a new empty database with the given ID.
	 * 
	 * @param ID
	 *            identification number of the created database
	 * @return the created database
	 * @throws IllegalArgumentException
	 *             if a database with the given ID already exists
	 * @throws SecurityException
	 *             if the platform cannot create dynamically databases
	 */
	public static ShieldedPlug createDatabase(int ID) {
		throw new SecurityException();
	}

	/**
	 * Gets if this database is immutable or not.
	 * 
	 * @return true if no block can be added or remove to this database, false otherwise
	 */
	public boolean isImmutable() {
		// do not need to ask native world
		return true;
	}

	/**
	 * Deletes the block with the given ID.
	 * 
	 * @param blockID
	 *            the ID of the block to delete
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws SecurityException
	 *             if this database is immutable
	 */
	public void delete(int blockID) {
		throw new SecurityException();
	}

	/**
	 * Creates a block with the given ID.
	 * 
	 * @param blockID
	 *            the ID of the block to create
	 * @param length
	 *            length in bytes of the block to create
	 * @param maxTasks
	 *            maximum number of tasks that can wait at the same time for a modification of the block
	 * @throws IllegalArgumentException
	 *             if a block is already defined with the given ID
	 * @throws SecurityException
	 *             if this database is immutable
	 */
	public void create(int blockID, int length, int maxTasks) {
		throw new SecurityException();
	}

	/**
	 * Creates a block with the given ID. An unlimited number of tasks will be able to wait at the same time for a modification of the block.
	 * 
	 * @param blockID
	 *            the ID of the block to create
	 * @param length
	 *            length in bytes of the block to create
	 * @throws IllegalArgumentException
	 *             if a block is already defined with the given ID
	 * @throws SecurityException
	 *             if this database is immutable
	 */
	public void create(int blockID, int length) {
		throw new SecurityException();
	}

	/**
	 * Gets the number of blocks of this database.
	 * 
	 * @return the number of blocks in this database
	 */
	public int getSize() {
		return NShieldedPlug.getSize(databasePtr);
	}

	/**
	 * Gets the ID of this database.
	 * 
	 * @return the ID of this database
	 */
	public int getID() {
		return NShieldedPlug.getID(databasePtr);
	}

	/**
	 * Gets the list of IDs of the blocks available in this database.
	 * 
	 * @return list of the IDs of the blocks available in this database
	 */
	public int[] getIDs() {
		int[] IDs = new int[getSize()];
		NShieldedPlug.getIDs(databasePtr, IDs);
		return IDs;
	}

	/**
	 * Returns the length of the block with the given ID.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return length in bytes
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public int getLength(int blockID) {
		int result = NShieldedPlug.getLength(databasePtr, blockID);
		checkBlockResult(result);
		return result;
	}

	/**
	 * Gets the maximum number of tasks that can wait at the same time on the block defined with the given ID.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the maximum number of tasks that can wait at the same time on the block defined with the given ID
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public int getMaxTasks(int blockID) {
		int result = NShieldedPlug.getMaxTasks(databasePtr, blockID);
		checkBlockResult(result);
		return result;
	}

	/**
	 * Fills the given array with data from the block with the given ID. The number of bytes read is equal to the length of <code>data</code>.<br>
	 * The <code>read(blockID, data)</code> method has the same effect as:
	 * 
	 * <pre>
	 * &lt;code&gt; read(blockID, data, 0, data.length) &lt;/code&gt;
	 * </pre>
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param data
	 *            the buffer into which the data is read
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws EmptyBlockException
	 *             if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 * 		if <code>data.length</code> is lower than block length
	 */
	public void read(int blockID, byte[] data) throws EmptyBlockException {
		int result = NShieldedPlug.read(databasePtr, blockID, data, 0, false);
		checkBlockReadResult(result);
	}

	/**
	 * Fills the given array with <code>block.length</code> bytes from the block with the given ID. The first byte read is stored into element <code>data[destOffset]</code>.
	 * <p>
	 * If <code>destOffset</code> is negative or <code>destOffset</code> + block length is greater than the length of the array <code>data</code>, then an <code>IndexOutOfBoundsException</code> is thrown.
	 * 
	 * @param blockID
	 *            the ID of the block
	 * @param data
	 *            the buffer into which the data is read
	 * @param destOffset
	 *            the start offset in array <code>data</code> at which the data is written
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws EmptyBlockException
	 *             if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 *             if <code>destOffset</code> is negative or if <code>data.length</code> is lower than <code>destOffset</code> + block length
	 */
	public void read(int blockID, byte[] data, int offset)
			throws EmptyBlockException {
		int result = NShieldedPlug.read(databasePtr, blockID, data, offset, false);
		checkBlockReadResult(result);
	}

	/**
	 * Reads four input bytes from the block with the given ID and returns an <code>int</code> value.<br>
	 * The way the <code>int</code> is built from the four bytes is platform dependent.<br>
	 * <p>
	 * This method is suitable for reading bytes written by the <code>writeInt</code> method.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the <code>int</code> value read
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws EmptyBlockException
	 *             if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 *             if block length is not four bytes
	 */
	public int readInt(int blockID) throws EmptyBlockException {
		byte[] bytes = new byte[INT_SIZE];
		int result = NShieldedPlug.read(databasePtr, blockID, bytes, 0, true);
		checkBlockReadResult(result);
		return ByteArray.readInt(bytes, 0);
	}

	/**
	 * Reads eight input bytes from the block with the given ID and returns a <code>long</code> value.<br>
	 * The way the <code>long</code> is built from the eight bytes is platform dependent.<br>
	 * <p>
	 * This method is suitable for reading bytes written by the <code>writeLong</code> method.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the <code>long</code> value read
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws EmptyBlockException
	 *             if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 *             if block length is not height bytes
	 */
	public long readLong(int blockID) throws EmptyBlockException {
		byte[] bytes = new byte[LONG_SIZE];
		int result = NShieldedPlug.read(databasePtr, blockID, bytes, 0, true);
		checkBlockReadResult(result);
		return ByteArray.readLong(bytes, 0);
	}

	/**
	 * Reads four input bytes from the block with the given ID and returns a <code>float</code> value.<br>
	 * The way the <code>float</code> is built from the four bytes is platform dependent.<br>
	 * <p>
	 * This method is suitable for reading bytes written by the <code>writeFloat</code> method.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the <code>float</code> value read
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws EmptyBlockException
	 *             if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 *             if block length is not four bytes
	 */
	public float readFloat(int blockID) throws EmptyBlockException {
		byte[] bytes = new byte[FLOAT_SIZE];
		int result = NShieldedPlug.read(databasePtr, blockID, bytes, 0, true);
		checkBlockReadResult(result);
		return Float.intBitsToFloat(ByteArray.readInt(bytes, 0));
	}

	/**
	 * Reads eight input bytes from the block with the given ID and returns a <code>double</code> value.<br>
	 * The way the <code>double</code> is built from the eight bytes is platform dependent.<br>
	 * <p>
	 * This method is suitable for reading bytes written by the <code>writeDouble</code> method.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the <code>double</code> value read
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws EmptyBlockException
	 *             if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 *             if block length is not height bytes
	 */
	public double readDouble(int blockID) throws EmptyBlockException {
		byte[] bytes = new byte[DOUBLE_SIZE];
		int result = NShieldedPlug.read(databasePtr, blockID, bytes, 0, true);
		checkBlockReadResult(result);
		return Double.longBitsToDouble(ByteArray.readLong(bytes, 0));
	}

	/**
	 * Invokes the <code>readObject</code> method of the {@link SPReader} registered for the block with the given ID.<br>
	 * The {@link SPReader} is responsible for the deserialization of the object from the block.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the object read from the block
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws EmptyBlockException
	 *             if no data is available in the block
	 * @throws NullPointerException
	 *             if no {@link SPReader} has been registered for the block
	 * @throws IndexOutOfBoundsException
	 *             if read data length is different than block length
	 */
	public Object readObject(int blockID) throws EmptyBlockException {
		return getReader(blockID).readObject(this, blockID);
	}

	/**
	 * Register the given {@link SPReader} to deserialize objects from the block with the given ID.<br>
	 * If an {@link SPReader} is already defined for the block, it is replaced by the given {@link SPReader}.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param reader
	 *            the {@link SPReader}
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public void setReader(int blockID, SPReader reader) {
		getBlock(blockID).reader = reader;
	}

	/**
	 * Gets the {@link SPReader} used to deserialize objects from the block with the given ID.<br>
	 * If no {@link SPReader} is defined for the block, <code>null</code> is returned.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the {@link SPReader} set or null if none
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public SPReader getReader(int blockID) {
		return getBlock(blockID).reader;
	}

	/**
	 * Writes <code>data.length</code> bytes from the specified byte array to the block with the given ID.<br>
	 * The <code>write(blockID, data)</code> method has the same effect as:
	 * 
	 * <pre>
	 * &lt;code&gt; write(blockID, data, 0, data.length) &lt;/code&gt;
	 * </pre>
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param data
	 *            the data to write
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 *             if <code>data.length</code> value is lower than block length
	 */
	public void write(int blockID, byte[] data) {
		int result = NShieldedPlug.write(databasePtr, blockID, data, 0, false);
		checkBlockResult(result);
	}

	/**
	 * Writes block length bytes from the specified byte array to the block with the given ID.<br>
	 * Element <code>data[destOffset]</code> is the first byte written to the block.
	 * <p>
	 * If <code>destOffset</code> is negative, or <code>destOffset</code> + block length is greater than the length of the array <code>data</code>, then an <code>IndexOutOfBoundsException</code> is thrown.
	 * 
	 * @param blockID
	 *            the ID of the block
	 * @param data
	 *            the data to write
	 * @param srcOffset
	 *            the start offset in the data
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 *             if <code>destOffset</code> is negative or if <code>data.length</code> is lower than <code>offset</code> + block length
	 */
	public void write(int blockID, byte[] data, int offset) {
		int result = NShieldedPlug.write(databasePtr, blockID, data, offset, false);
		checkBlockResult(result);
	}

	/**
	 * Writes an <code>int</code> value, which is comprised of four bytes, to the block with the given ID.<br>
	 * The way the <code>int</code> is written from the four bytes is platform dependent.<br>
	 * <p>
	 * The bytes written by this method may be read by the <code>readInt</code> method, which will then return an <code>int</code> equal to <code>value</code>.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param value
	 *            the <code>int</code> value to be written
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 *             if block length is not four bytes
	 */
	public void writeInt(int blockID, int value) {
		byte[] bytes = new byte[INT_SIZE];
		ByteArray.writeInt(bytes, 0, value);
		int result = NShieldedPlug.write(databasePtr, blockID, bytes, 0, true);
		checkBlockResult(result);
	}

	/**
	 * Writes a <code>long</code> value, which is comprised of eight bytes, to the block with the given ID.<br>
	 * The way the <code>long</code> is written from the eight bytes is platform dependent.<br>
	 * <p>
	 * The bytes written by this method may be read by the <code>readLong</code> method, which will then return a <code>long</code> equal to <code>value</code>.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param value
	 *            the <code>long</code> value to be written
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 *             if block length is not eight bytes
	 */
	public void writeLong(int blockID, long value) {
		byte[] bytes = new byte[LONG_SIZE];
		ByteArray.writeLong(bytes, 0, value);
		int result = NShieldedPlug.write(databasePtr, blockID, bytes, 0, true);
		checkBlockResult(result);
	}

	/**
	 * Writes a <code>float</code> value, which is comprised of four bytes, to the block with the given ID.<br>
	 * The way the <code>float</code> is written from the four bytes is platform dependent.<br>
	 * <p>
	 * The bytes written by this method may be read by the <code>readFloat</code> method, which will then return a <code>float</code> equal to <code>value</code>.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param value
	 *            the <code>float</code> value to be written
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 *             if block length is not four bytes
	 */
	public void writeFloat(int blockID, float value) {
		byte[] bytes = new byte[FLOAT_SIZE];
		ByteArray.writeInt(bytes, 0, Float.floatToIntBits(value));
		int result = NShieldedPlug.write(databasePtr, blockID, bytes, 0, true);
		checkBlockResult(result);
	}

	/**
	 * Writes a <code>double</code> value, which is comprised of eight bytes, to the block with the given ID.<br>
	 * The way the <code>double</code> is written from the eight bytes is platform dependent.<br>
	 * <p>
	 * The bytes written by this method may be read by the <code>readDouble</code> method, which will then return a <code>double</code> equal to <code>value</code>.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param value
	 *            the <code>double</code> value to be written
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 *             if block length is not eight bytes
	 */
	public void writeDouble(int blockID, double value) {
		byte[] bytes = new byte[DOUBLE_SIZE];
		ByteArray.writeLong(bytes, 0, Double.doubleToLongBits(value));
		int result = NShieldedPlug.write(databasePtr, blockID, bytes, 0, true);
		checkBlockResult(result);
	}

	/**
	 * Invokes the <code>writeObject</code> method of the {@link SPWriter} registered for the block with the given ID.<br>
	 * The {@link SPWriter} is responsible for the serialization of the object into the block.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param o
	 *            the object to be written
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws NullPointerException
	 *             if no {@link SPWriter} has been registered for the block
	 * @throws IndexOutOfBoundsException
	 *             if block length is lower than the size needed for object serialization
	 */
	public void writeObject(int blockID, Object o) {
		getWriter(blockID).writeObject(this, blockID, o);
	}

	/**
	 * Register the given {@link SPWriter} to serialize objects into the block with the given ID.<br>
	 * If an {@link SPWriter} is already defined for the block, it is replaced by the given {@link SPWriter}.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @param reader
	 *            the {@link SPWriter}
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public void setWriter(int blockID, SPWriter writer) {
		getBlock(blockID).writer = writer;
	}

	/**
	 * Gets the {@link SPWriter} used to serialize objects into the block with the given ID.<br>
	 * If no {@link SPWriter} is defined for the block, <code>null</code> is returned.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return the {@link SPWriter} set or null if none
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public SPWriter getWriter(int blockID) {
		return getBlock(blockID).writer;
	}

	/**
	 * Causes current thread to wait until another thread write data into the block with the given ID.<br>
	 * If data has been written in the block since last read, this method returns immediately.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @throws InterruptedException
	 *             if another thread has interrupted the current thread The interrupted status of the current thread is cleared when this exception is thrown
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 * @throws IllegalStateException
	 *             if too many threads are waiting for new data
	 */
	public void waitFor(int blockID) throws InterruptedException {
		try {
			int result = NShieldedPlug.waitOne(databasePtr, blockID);
			checkBlockResult(result);
			if (result == NShieldedPlug.WAITING_FOR) {
				NShieldedPlug.removeWaitFlag(databasePtr, blockID);
			}
		} catch (InterruptedException e) {
			NShieldedPlug.removeWaitFlag(databasePtr, blockID);
			throw new InterruptedException();
		}
	}

	/**
	 * Causes current thread to wait until another thread write data into at least one block from the blocks with the given IDs.<br>
	 * If data has been written in one block since last read from it, this method returns immediately.
	 * 
	 * @param blockIDs
	 *            list of block IDs
	 * @return list of IDs of the blocks that has been written
	 * @throws InterruptedException
	 *             if another thread has interrupted the current thread The interrupted status of the current thread is cleared when this exception is thrown
	 * @throws IllegalArgumentException
	 *             if one of the ID does not correspond to an existing block
	 * @throws IllegalStateException
	 *             if too many threads are waiting for new data
	 */
	public int[] waitFor(int[] blockIDs) throws InterruptedException {
		int length = blockIDs.length;
		boolean[] writtenBlocks = new boolean[length];
		try {
			int result = NShieldedPlug.waitSeveral(databasePtr, blockIDs);
			checkBlockResult(result);
			// result == NShieldedPlug.SUCCESS for sure
			NShieldedPlug.removeWaitFlags(databasePtr, blockIDs, writtenBlocks);
			int[] writtenBlocksIDs = new int[length];
			int writtenBlocksCpt = 0;
			for (int i = -1; ++i < length;) {
				if (writtenBlocks[i]) {
					writtenBlocksIDs[writtenBlocksCpt++] = blockIDs[i];
				}
			}
			System.arraycopy(writtenBlocksIDs, 0, writtenBlocksIDs = new int[writtenBlocksCpt], 0, writtenBlocksCpt);
			return writtenBlocksIDs;
		} catch (InterruptedException e) {
			NShieldedPlug.removeWaitFlags(databasePtr, blockIDs, writtenBlocks);
			throw new InterruptedException();
		}
	}

	/**
	 * Gets if data has been written into the block since last read.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return <code>true</code> if data has been written into the block since last read, <code>false</code> otherwise
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public boolean isPending(int blockID) {
		return getState(databasePtr, blockID, NShieldedPlug.PENDING);
	}

	/**
	 * Determines whether data in the block with the given ID are available or not.<br>
	 * By default no data is available in a block. When data are written in a block, they remain available until method <code>reset(int)</code> is called.
	 * 
	 * @param blockID
	 *            ID of the block
	 * @return <code>true</code> if data is available in the block <code>false</code> otherwise
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public boolean isDataAvailable(int blockID) {
		return !getState(databasePtr, blockID, NShieldedPlug.DISABLE);
	}

	/**
	 * Resets data of the block with the given ID.<br>
	 * After execution of this method, <code>isDataAvailable(int)</code> method would return <code>false</code> (unless data were written after calling <code>reset(int)</code> and before calling {@link #isDataAvailable(int)}).
	 * 
	 * @param blockID
	 *            ID of the block
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	public void reset(int blockID) {
		NShieldedPlug.reset(databasePtr, blockID);
	}

	/**
	 * Check the result of a block native.
	 * 
	 * @param result
	 *            the result to check
	 * @throws IllegalArgumentException
	 *             if result == {@link NShieldedPlug#NO_BLOCK}
	 * @throws IndexOutOfBoundsException
	 *             if result == {@link NShieldedPlug#INDEX_OUT_OF_BOUNDS}
	 * @throws TooManyWaitingThreadsException
	 *             if result == {@link NShieldedPlug#TOO_MANY_WAITING}
	 */
	private static void checkBlockResult(int result) {
		switch (result) {
			case NShieldedPlug.NO_BLOCK:
				throw new IllegalArgumentException();
			case NShieldedPlug.INDEX_OUT_OF_BOUNDS:
				throw new IndexOutOfBoundsException();
			case NShieldedPlug.TOO_MANY_WAITING:
				throw new TooManyWaitingThreadsException();
		}
	}

	/**
	 * Check the result of a read block native.
	 * 
	 * @param result
	 *            the result to check
	 * @throws IllegalArgumentException
	 *             if result == {@link NShieldedPlug#NO_BLOCK}
	 * @throws EmptyBlockException
	 *             if result == {@link NShieldedPlug#EMPTY_BLOCK}
	 * @throws IndexOutOfBoundsException
	 *             if result == {@link NShieldedPlug#INDEX_OUT_OF_BOUNDS}
	 */
	private static void checkBlockReadResult(int result)
			throws EmptyBlockException {
		switch (result) {
		case NShieldedPlug.NO_BLOCK:
			throw new IllegalArgumentException();
		case NShieldedPlug.EMPTY_BLOCK:
			throw new EmptyBlockException();
		case NShieldedPlug.INDEX_OUT_OF_BOUNDS:
			throw new IndexOutOfBoundsException();
		}
	}

	/**
	 * Check if the state of the block match the given flag.
	 * 
	 * @param databasePtr
	 *            the database
	 * @param blockID
	 *            the block ID
	 * @param flag
	 *            the flag to compare with
	 * @return <code>true</code> if <code>state == flag</code>, <code>false</code> otherwise.
	 * @throws IllegalArgumentException
	 *             if no block is defined with the given ID
	 */
	private static boolean getState(int databasePtr, int blockID, int flag) {
		int result = NShieldedPlug.getDataState(databasePtr, blockID);
		checkBlockResult(result);
		return result == flag;
	}

	/**
	 * Returns the block with the given ID. Create it if not already exists.
	 * 
	 * @param blockID
	 *            the ID of the block
	 * @return the block
	 */
	private Block getBlock(int blockID) {
		checkBlockResult(NShieldedPlug.exists(databasePtr, blockID));
		synchronized (blocks) {
			Block block = (Block) blocks.at(blockID);
			if (block == null) {
				// not found, get from native
				block = new Block();
				blocks.atPutAlreadyCheckNotExists(blockID, block);
			}
			return block;
		}
	}

}
