/**
 * Java
 *
 * Copyright 2011-2015 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.sp;

import com.is2t.sp.remote.RemoteShieldedPlug;
import com.is2t.sp.remote.RemoteShieldedPlugConnector;
import com.is2t.sp.remote.request.ReqCreateBlock;
import com.is2t.sp.remote.request.ReqCreateDatabase;
import com.is2t.sp.remote.request.ReqDeleteBlock;
import com.is2t.sp.remote.request.ReqGetIDs;
import com.is2t.sp.remote.request.ReqGetLength;
import com.is2t.sp.remote.request.ReqGetMaxTasks;
import com.is2t.sp.remote.request.ReqGetSize;
import com.is2t.sp.remote.request.ReqIsBigEndian;
import com.is2t.sp.remote.request.ReqIsDataAvailable;
import com.is2t.sp.remote.request.ReqIsImmutable;
import com.is2t.sp.remote.request.ReqIsPending;
import com.is2t.sp.remote.request.ReqIsValidDatabaseId;
import com.is2t.sp.remote.request.ReqReadArray;
import com.is2t.sp.remote.request.ReqReadInt;
import com.is2t.sp.remote.request.ReqReadLong;
import com.is2t.sp.remote.request.ReqReset;
import com.is2t.sp.remote.request.ReqWaitFor;
import com.is2t.sp.remote.request.ReqWriteArray;
import com.is2t.sp.remote.request.ReqWriteInt;
import com.is2t.sp.remote.request.ReqWriteLong;


public class ShieldedPlug {
	//lazy creation when calling static method in this class
	private static RemoteShieldedPlug RemoteSP;
	private static boolean isBigEndian;

	private static final int DEFAULT_SP_POOL_SIZE = 4;
	private static ShieldedPlug[] ShieldedPlugs = new ShieldedPlug[DEFAULT_SP_POOL_SIZE];
	private static int ShieldedPlugsPtr = -1;

	private int databaseID;

	private class SPBlockReader {
		public int blockID;
		public SPReader spReader;

		public SPBlockReader(int blockID, SPReader spReader) {
			this.blockID = blockID;
			this.spReader = spReader;
		}
	}

	private SPBlockReader[] spBlockReaders;

	private class SPBlockWriter {
		public int blockID;
		public SPWriter spWriter;

		public SPBlockWriter(int blockID, SPWriter spWriter) {
			this.blockID = blockID;
			this.spWriter = spWriter;
		}
	}

	private SPBlockWriter[] spBlockWriters;

	/**
	 * Could not create a <code>ShieldedPlug</code>
	 */
	private ShieldedPlug(int databaseID){
		this.databaseID = databaseID;
		this.spBlockReaders = new SPBlockReader[0];
		this.spBlockWriters = new SPBlockWriter[0];
	}

	/**
	 * Returns the database defined at the given ID.
	 *
	 * @param ID
	 * 		identification number of the requested database
	 * @return
	 * 		the database with the given ID
	 * @throws IllegalArgumentException
	 * 		if no database is defined with the given ID
	 */
	synchronized public static ShieldedPlug getDatabase(int ID){
		initialize();

		//check if SP is already in the pool
		for(int i=ShieldedPlugsPtr+1 ; --i>=0 ;){
			if(ShieldedPlugs[i].databaseID == ID)
				return ShieldedPlugs[i];
		}

		//SP not found in the pool, create a new one
		RemoteSP.executeRequest(new ReqIsValidDatabaseId(ID));//may throw an exception

		//create a new SP and add it in the pool
		ShieldedPlug sp = new ShieldedPlug(ID);
		try{
			ShieldedPlugs[++ShieldedPlugsPtr] = sp;
		}
		catch(ArrayIndexOutOfBoundsException e){
			//array too small
			System.arraycopy(ShieldedPlugs, 0, ShieldedPlugs=new ShieldedPlug[ShieldedPlugsPtr*2], 0, ShieldedPlugsPtr);
			ShieldedPlugs[ShieldedPlugsPtr] = sp;
		}

		return sp;
	}

	/**
	 * Creates a new empty database with the given ID.
	 *
	 * @param ID
	 * 		identification number of the created database
	 * @return
	 * 		the created database
	 * @throws IllegalArgumentException
	 * 		if a database with the given ID already exists
	 * @throws SecurityException
	 * 		if the platform cannot create dynamically databases
	 */
	synchronized public static ShieldedPlug createDatabase(int ID){
		initialize();

		//throws the corresponding exception if an error occured
		RemoteSP.executeRequest(new ReqCreateDatabase(ID));

		return getDatabase(ID);
	}

	/**
	 * Gets if this database is immutable or not.
	 *
	 * @return
	 * 		true if no block can be added or remove to this database, false otherwise
	 */
	public boolean isImmutable(){
		return ((Boolean)RemoteSP.executeRequest(new ReqIsImmutable(this))).booleanValue();
	}

	/**
	 * Deletes the block with the given ID.
	 *
	 * @param blockID
	 * 		the ID of the block to delete
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws SecurityException
	 * 		if this database is immutable
	 */
	public void delete(int blockID){
		RemoteSP.executeRequest(new ReqDeleteBlock(this, blockID));
	}

	/**
	 * Creates a block with the given ID.
	 *
	 * @param blockID
	 * 		the ID of the block to create
	 * @param length
	 * 		the length in bytes of the block to create
	 * @param maxTasks
	 * 		maximum number of tasks that can wait at the same time for a modification of the block
	 * @throws IllegalArgumentException
	 * 		if a block is already defined with the given ID
	 * @throws SecurityException
	 * 		if this database is immutable
	 */
	public void create(int blockID, int length, int maxTasks){
		RemoteSP.executeRequest(new ReqCreateBlock(this, blockID, length, maxTasks));
	}

	/**
	 * Creates a block with the given ID. An unlimited number of tasks will be able to wait at the
	 * same time for a modification of the block.
	 *
	 * @param blockID
	 * 		the ID of the block to create
	 * @param length
	 * 		the length in bytes of the block to create
	 * @throws IllegalArgumentException
	 * 		if a block is already defined with the given ID
	 * @throws SecurityException
	 * 		if this database is immutable
	 */
	public void create(int blockID, int length){
		create(blockID, length, -1);
	}

	/**
	 * Gets the number of blocks of this database.
	 *
	 * @return
	 * 		the number of blocks in this database
	 */
	public int getSize(){
		return ((Integer)RemoteSP.executeRequest(new ReqGetSize(this))).intValue();
	}

	/**
	 * Gets the ID of this database.
	 *
	 * @return
	 * 		the ID of this database
	 */
	public int getID(){
		return databaseID;
	}

	/**
	 * Gets the list of IDs of the blocks available in this database.
	 *
	 * @return
	 * 		list of the IDs of the blocks available in this database
	 */
	public int[] getIDs(){
		return (int[])RemoteSP.executeRequest(new ReqGetIDs(this));
	}

	/**
	 * Returns the length of the block with the given ID.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		length in bytes
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public int getLength(int blockID){
		return ((Integer)RemoteSP.executeRequest(new ReqGetLength(this, blockID))).intValue();
	}

	/**
	 * Gets the maximum number of tasks that can wait at the same time on the block defined with the given ID.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the maximum number of tasks that can wait at the same time on the block defined with the given ID
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public int getMaxTasks(int blockID){
		return ((Integer)RemoteSP.executeRequest(new ReqGetMaxTasks(this, blockID))).intValue();
	}

	/**
	 * Fills the given array with data from the block with the given ID.
	 * The number of bytes read is equal to the length of the block.<br>
	 * The <code>read(blockID, data)</code> method has the same effect as:
	 * <pre><code> read(blockID, data, 0) </code></pre>
	 *
	 * @param blockID
	 * 		the ID of the block
	 * @param data
	 * 		the buffer into which the data is read
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws EmptyBlockException
	 * 		if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 * 		if <code>data.length</code> is lower than block length
	 */
	public void read(int blockID, byte[] data) throws EmptyBlockException{
		read(blockID, data, 0);
	}

	/**
	 * Fills the given array with <code>block.length</code> bytes from the block with the given ID.
	 * The first byte read is stored into element <code>data[destOffset]</code>.
	 * <p> If <code>destOffset</code> is negative or <code>destOffset</code> + block length
	 * is greater than the length of the array <code>data</code>, then an
	 * <code>IndexOutOfBoundsException</code> is thrown.
	 *
	 * @param blockID
	 * 		the ID of the block
	 * @param data
	 * 		the buffer into which the data is read
	 * @param destOffset
	 *		the start offset in array <code>data</code> at which the data is written
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws EmptyBlockException
	 * 		if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 * 		if <code>destOffset</code> is negative or
	 * 		if <code>data.length</code> is lower than <code>destOffset</code> + block length
	 */
	public void read(int blockID, byte[] data, int offset) throws EmptyBlockException{
		RemoteSP.executeReadBlockRequest(new ReqReadArray(this, blockID, data, offset));
	}

	/**
	 * Reads four input bytes from the block with the given ID and returns an <code>int</code> value.<br>
	 * The way the <code>int</code> is built from the four bytes is platform dependent.<br>
	 * <p>This method is suitable for reading bytes written by the <code>writeInt</code> method.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the <code>int</code> value read
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws EmptyBlockException
	 * 		if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than four bytes
	 */
	public int readInt(int blockID) throws EmptyBlockException{
		return ((Integer)RemoteSP.executeReadBlockRequest(new ReqReadInt(this, blockID))).intValue();
	}

	/**
	 * Reads eight input bytes from the block with the given ID and returns a <code>long</code> value.<br>
	 * The way the <code>long</code> is built from the eight bytes is platform dependent.<br>
	 * <p>This method is suitable for reading bytes written by the <code>writeLong</code> method.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the <code>long</code> value read
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws EmptyBlockException
	 * 		if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than height bytes
	 */
	public long readLong(int blockID) throws EmptyBlockException{
		return ((Long)RemoteSP.executeReadBlockRequest(new ReqReadLong(this, blockID))).longValue();
	}

	/**
	 * Reads four input bytes from the block with the given ID and returns a <code>float</code> value.<br>
	 * The way the <code>float</code> is built from the four bytes is platform dependent.<br>
	 * <p>This method is suitable for reading bytes written by the <code>writeFloat</code> method.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the <code>float</code> value read
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws EmptyBlockException
	 * 		if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than four bytes
	 */
	public float readFloat(int blockID) throws EmptyBlockException{
		return Float.intBitsToFloat(readInt(blockID));
	}

	/**
	 * Reads eight input bytes from the block with the given ID and returns a <code>double</code> value.<br>
	 * The way the <code>double</code> is built from the eight bytes is platform dependent.<br>
	 * <p>This method is suitable for reading bytes written by the <code>writeDouble</code> method.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the <code>double</code> value read
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws EmptyBlockException
	 * 		if no data is available in the block
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than height bytes
	 */
	public double readDouble(int blockID) throws EmptyBlockException{
		return Double.longBitsToDouble(readLong(blockID));
	}

	/**
	 * Invokes the <code>readObject</code> method of the {@link SPReader} registered for the block
	 * with the given ID.<br>
	 * The {@link SPReader} is responsible for the deserialization of the object from the block.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the object read from the block
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws EmptyBlockException
	 * 		if no data is available in the block
	 * @throws NullPointerException
	 * 		if no {@link SPReader} has been registered for the block
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than the size needed for object deserialization
	 */
	public Object readObject(int blockID) throws EmptyBlockException{
		return getReader(blockID).readObject(this, blockID);
	}

	/**
	 * Register the given {@link SPReader} to deserialize objects from the block with the given ID.<br>
	 * If an {@link SPReader} is already defined for the block, it is replaced by the given {@link SPReader}.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @param reader
	 * 		the {@link SPReader}
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public void setReader(int blockID, SPReader reader){
		// call a method to throw an exception if the block doesn't exists
		isDataAvailable(blockID);

		synchronized (spBlockReaders) {
			// replace the reader if it already exists
			int spBlockReadersLength = spBlockReaders.length;
			for (int i = -1; ++i < spBlockReadersLength;) {
				SPBlockReader spBlockReader = spBlockReaders[i];

				if (spBlockReader.blockID == blockID) {
					spBlockReader.spReader = reader;
					return;
				}
			}

			// else add it to the array
			SPBlockReader[] tmpSPBlockReaders = new SPBlockReader[spBlockReaders.length + 1];
			System.arraycopy(spBlockReaders, 0, tmpSPBlockReaders, 0, spBlockReaders.length);
			tmpSPBlockReaders[spBlockReaders.length] = new SPBlockReader(blockID, reader);
			spBlockReaders = tmpSPBlockReaders;
		}
	}

	/**
	 * Gets the {@link SPReader} used to deserialize objects from the block with the given ID.<br>
	 * If no {@link SPReader} is defined for the block, <code>null</code> is returned.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the {@link SPReader} set or null if none
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public SPReader getReader(int blockID){
		// call a method to throw an exception if the block doesn't exists
		isDataAvailable(blockID);

		synchronized (spBlockReaders) {
			int spBlockReadersLength = spBlockReaders.length;
			for (int i = -1; ++i < spBlockReadersLength;) {
				SPBlockReader spBlockReader = spBlockReaders[i];

				if (spBlockReader.blockID == blockID)
					return spBlockReader.spReader;
			}
		}

		return null;
	}

	/**
	 * Writes block length bytes from the specified byte array to the block
	 * with the given ID.<br>
	 * The <code>write(blockID, data)</code> method has the same effect as:
	 * <pre><code> write(blockID, data, 0) </code></pre>
	 *
	 * @param blockID
	 * 		the ID of the block
	 * @param data
	 * 		the data to write
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 * 		if <code>data.length</code> value is higher than block length
	 */
	public void write(int blockID, byte[] data){
		write(blockID, data, 0);
	}

	/**
	 * Writes block length bytes from the specified byte array to the block
	 * with the given ID.<br>
	 * Element <code>data[destOffset]</code> is the first byte written to the block.
	 * <p>
     * If <code>destOffset</code> is negative, or
     * <code>destOffset</code> + block length is greater than the length of the array
     * <code>data</code>, then an <code>IndexOutOfBoundsException</code> is thrown.
	 *
	 * @param blockID
	 * 		the ID of the block
	 * @param data
	 * 		the data to write
	 * @param destOffset
	 * 		the start offset in the data
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 * 		if <code>destOffset</code> is negative or
	 * 		if <code>data.length</code> is lower than <code>offset</code> + block length
	 */
	public void write(int blockID, byte[] data, int offset){
		RemoteSP.executeRequest(new ReqWriteArray(this, blockID, data, offset));
	}

	/**
	 * Writes an <code>int</code> value, which is comprised of four bytes, to the block with the given ID.<br>
	 * The way the <code>int</code> is written from the four bytes is platform dependent.<br>
	 * <p>The bytes written by this method may be read by the <code>readInt</code> method, which will then
     * return an <code>int</code> equal to <code>value</code>.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @param value
	 * 		the <code>int</code> value to be written
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than four bytes
	 */
	public void writeInt(int blockID, int value){
		RemoteSP.executeRequest(new ReqWriteInt(this, blockID, value));
	}

	/**
	 * Writes a <code>long</code> value, which is comprised of eight bytes, to the block with the given ID.<br>
	 * The way the <code>long</code> is written from the eight bytes is platform dependent.<br>
	 * <p>The bytes written by this method may be read by the <code>readLong</code> method, which will then
     * return a <code>long</code> equal to <code>value</code>.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @param value
	 * 		the <code>long</code> value to be written
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than height bytes
	 */
	public void writeLong(int blockID, long value){
		RemoteSP.executeRequest(new ReqWriteLong(this, blockID, value));
	}

	/**
	 * Writes a <code>float</code> value, which is comprised of four bytes, to the block with the given ID.<br>
	 * The way the <code>float</code> is written from the four bytes is platform dependent.<br>
	 * <p>The bytes written by this method may be read by the <code>readFloat</code> method, which will then
     * return a <code>float</code> equal to <code>value</code>.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @param value
	 * 		the <code>float</code> value to be written
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than four bytes
	 */
	public void writeFloat(int blockID, float value){
		writeInt(blockID, Float.floatToIntBits(value));
	}

	/**
	 * Writes a <code>double</code> value, which is comprised of eight bytes, to the block with the given ID.<br>
	 * The way the <code>double</code> is written from the eight bytes is platform dependent.<br>
	 * <p>The bytes written by this method may be read by the <code>readDouble</code> method, which will then
     * return a <code>double</code> equal to <code>value</code>.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @param value
	 * 		the <code>double</code> value to be written
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than eight bytes
	 */
	public void writeDouble(int blockID, double value){
		writeLong(blockID, Double.doubleToLongBits(value));
	}

	/**
	 * Invokes the <code>writeObject</code> method of the {@link SPWriter} registered for the block
	 * with the given ID.<br>
	 * The {@link SPWriter} is responsible for the serialization of the object into the block.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @param o
	 * 		the object to be written
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws NullPointerException
	 * 		if no {@link SPWriter} has been registered for the block
	 * @throws IndexOutOfBoundsException
	 * 		if block length is lower than the size needed for object serialization
	 */
	public void writeObject(int blockID, Object o){
		getWriter(blockID).writeObject(this, blockID, o);
	}

	/**
	 * Register the given {@link SPWriter} to serialize objects into the block with the given ID.<br>
	 * If an {@link SPWriter} is already defined for the block, it is replaced by the given {@link SPWriter}.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @param reader
	 * 		the {@link SPWriter}
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public void setWriter(int blockID, SPWriter writer){
		// call a method to throw an exception if the block doesn't exists
		isDataAvailable(blockID);

		synchronized (spBlockWriters) {
			// replace the reader if it already exists
			int spBlockWritersLength = spBlockWriters.length;
			for (int i = -1; ++i < spBlockWritersLength;) {
				SPBlockWriter spBlockWriter = spBlockWriters[i];

				if (spBlockWriter.blockID == blockID) {
					spBlockWriter.spWriter = writer;
					return;
				}
			}

			// else add it to the array
			SPBlockWriter[] tmpSPBlockWriters = new SPBlockWriter[spBlockWriters.length + 1];
			System.arraycopy(spBlockWriters, 0, tmpSPBlockWriters, 0, spBlockWriters.length);
			tmpSPBlockWriters[spBlockWriters.length] = new SPBlockWriter(blockID, writer);
			spBlockWriters = tmpSPBlockWriters;
		}
	}

	/**
	 * Gets the {@link SPWriter} used to serialize objects into the block with the given ID.<br>
	 * If no {@link SPWriter} is defined for the block, <code>null</code> is returned.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		the {@link SPWriter} set or null if none
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public SPWriter getWriter(int blockID){
		// call a method to throw an exception if the block doesn't exists
		isDataAvailable(blockID);

		synchronized (spBlockWriters) {
			int spBlockWritersLength = spBlockWriters.length;
			for (int i = -1; ++i < spBlockWritersLength;) {
				SPBlockWriter spBlockWriter = spBlockWriters[i];

				if (spBlockWriter.blockID == blockID)
					return spBlockWriter.spWriter;
			}
		}

		return null;
	}

	/**
	 * Causes current thread to wait until another thread write data into the block
	 * with the given ID.<br>
	 * If data has been written in the block since last read, this method
	 * returns immediately.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @throws InterruptedException
	 * 		if another thread has interrupted the current thread
	 * 		The interrupted status of the current thread is cleared
	 * 		when this exception is thrown
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 * @throws TooManyWaitingThreadsException
	 * 		if too many threads are waiting for new data
	 */
	public void waitFor(int blockID) throws InterruptedException{
		waitFor(new int[]{blockID});
	}

	/**
	 * Causes current thread to wait until another thread write data into at least one block
	 * from the blocks with the given IDs.<br>
	 * If data has been written in one block since last read from it, this method
	 * returns immediately.
	 *
	 * @param blockIDs
	 * 		list of block IDs
	 * @return
	 * 		list of IDs of the blocks that has been written
	 * @throws InterruptedException
	 * 		if another thread has interrupted the current thread
	 * 		The interrupted status of the current thread is cleared
	 * 		when this exception is thrown
	 * @throws IllegalArgumentException
	 * 		if one of the ID does not correspond to an existing block
	 * @throws TooManyWaitingThreadsException
	 * 		if too many threads are waiting for new data
	 */
	public int[] waitFor(int[] blockIDs) throws InterruptedException{
		return (int[])RemoteSP.executeInterruptableRequest(new ReqWaitFor(this, blockIDs));
	}

	/**
	 * Gets if data has been written into the block since last read.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		<code>true</code> if data has been written into the block since last
	 * 		read, <code>false</code> otherwise
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public boolean isPending(int blockID){
		return ((Boolean)RemoteSP.executeRequest(new ReqIsPending(this, blockID))).booleanValue();
	}

	/**
	 * Determines whether data in the block with the given ID are available or not.<br>
	 * By default no data is available in a block. When data are written in a block, they remain
	 * available until method <code>reset(int)</code> is called.
	 *
	 * @param blockID
	 * 		ID of the block
	 * @return
	 * 		<code>true</code> if data is available in the block
	 * 		<code>false</code> otherwise
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public boolean isDataAvailable(int blockID){
		return ((Boolean)RemoteSP.executeRequest(new ReqIsDataAvailable(this, blockID))).booleanValue();
	}

	/**
	 * Resets data of the block with the given ID.<br>
	 * After execution of this method, {@link #isDataAvailable(int)}
	 * method would return <code>false</code> (unless data were written after calling <code>reset(int)</code> and before
	 * calling {@link #isDataAvailable(int)}).
	 *
	 * @param blockID
	 *		ID of the block
	 * @throws IllegalArgumentException
	 * 		if no block is defined with the given ID
	 */
	public void reset(int blockID){
		RemoteSP.executeRequest(new ReqReset(this, blockID));
	}

	private static void initialize(){
		if (RemoteSP != null) {
			return;
		}
		ShieldedPlugs = new ShieldedPlug[DEFAULT_SP_POOL_SIZE];
		RemoteShieldedPlugConnector connector;
		try {
			connector = getConnector();
		} catch (ClassNotFoundException e) {
			throw new RuntimeException(e.toString());
		}
		RemoteSP = new RemoteShieldedPlug(connector);
		// get the endianess of the remote host
		isBigEndian = ((Boolean) RemoteSP.executeRequest(new ReqIsBigEndian())).booleanValue();
	}

	private static RemoteShieldedPlugConnector getConnector() throws ClassNotFoundException {
		String name = System.getProperty(RemoteShieldedPlugConnector.CONNECTOR_ID);
		try {
			return (RemoteShieldedPlugConnector) Class.forName(name).newInstance();
		} catch (Throwable e) {
			throw new ClassNotFoundException(e.toString());
		}
	}

}
