/*
 * Copyright (c) 2000, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2016, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package java.net;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.NoSuchElementException;

import com.is2t.support.DependencyInjectionHelper;
import com.is2t.support.net.INetworkInterface;
import com.is2t.support.net.INetworkInterfaceData;
import com.is2t.support.net.security.IURLSecurityManager;

public final class NetworkInterface {
	private final String name;
	private final String displayName;
	private final InetAddress addrs[];
	private final short prefixLengthes[];
	private final InetAddress broadcasts[];

	// CLDC-1.1-COMPLIANCE
	private static INetworkInterface VMNetworkInterface = (INetworkInterface) DependencyInjectionHelper.getImplementation(INetworkInterface.class);
	private static IURLSecurityManager URLSecurityManager = (IURLSecurityManager) DependencyInjectionHelper.getImplementation(IURLSecurityManager.class);

	NetworkInterface(INetworkInterfaceData data) throws SocketException {
		this.name = data.getName();
		
		byte[][] netifDataAddrs = data.getAddresses();
		if ((netifDataAddrs != null) && (netifDataAddrs.length > 0)) {
			addrs = buildAddresses(netifDataAddrs);
		} else {
			addrs = null;
		}
		
		prefixLengthes = data.getPrefixes();
		
		byte[][] broadcastAddrs = data.getBroadcasts();
		if ((broadcastAddrs != null) && (broadcastAddrs.length > 0)) {
			broadcasts = buildAddresses(broadcastAddrs);
		} else {
			broadcasts = null;
		}
		
		// FIXME: that's weird, this field is never used
		displayName = null;
	}

	private static InetAddress[] buildAddresses(byte[][] addrs) throws SocketException {
		InetAddress[] netifAddrs = new InetAddress[addrs.length];
		for(int j=0; j < addrs.length; j++){
			if (addrs[j] != null) {
				try {
					netifAddrs[j] = InetAddress.getByAddress(addrs[j]);
				} catch (UnknownHostException e) {
					SocketException se = new SocketException();
					se.initCause(e);
					throw se;
				}
			}
		}
		return netifAddrs;
	}

	public String getName() {
		return name;
	}

	public Enumeration<InetAddress> getInetAddresses() {

		class checkedAddresses implements Enumeration<InetAddress> {

			private int i=0, count=0;
            private final InetAddress local_addrs[];

			checkedAddresses() {
				if (addrs != null) {
					local_addrs = new InetAddress[addrs.length];
					boolean trusted = true;
	
					SecurityManager sec = System.getSecurityManager();
					if (sec != null) {
						try {
							URLSecurityManager.checkGetNetworkInformation(sec);
						} catch (SecurityException e) {
							trusted = false;
						}
					}
					for (int j=0; j<addrs.length; j++) {
						try {
							if (sec != null && !trusted) {
								URLSecurityManager.checkConnect(sec, addrs[j].getHostAddress(), -1);
							}
							local_addrs[count++] = addrs[j];
						} catch (SecurityException e) { }
					}
				} else {
					local_addrs = new InetAddress[0];
				}
			}

			public InetAddress nextElement() {
				if (i < count) {
					return local_addrs[i++];
				} else {
					throw new NoSuchElementException();
				}
			}

			public boolean hasMoreElements() {
				return (i < count);
			}
		}
		return new checkedAddresses();

	}

    public java.util.List<InterfaceAddress> getInterfaceAddresses() {
    	int len;
    	if (addrs != null) {
    		len = addrs.length;
    	} else {
    		len = 0;
    	}
    	List<InterfaceAddress> list = new ArrayList<InterfaceAddress>(len);
		boolean trusted = true;

		SecurityManager sec = System.getSecurityManager();
		if (sec != null) {
			try {
				URLSecurityManager.checkGetNetworkInformation(sec);
			} catch (SecurityException e) {
				trusted = false;
			}
		}
		for (int j=0; j<len; j++) {
			try {
				if (sec != null && !trusted) {
					URLSecurityManager.checkConnect(sec, addrs[j].getHostAddress(), -1);
				}
				
				InterfaceAddress a = new InterfaceAddress(addrs[j], prefixLengthes[j], broadcasts[j]);
				list.add(a);
			} catch (SecurityException e) { }
		}
        return list;
    }

    public int getIndex() {
		try {
			INetworkInterfaceData[] netifDatas = VMNetworkInterface.getVMInterfaces();

			if (netifDatas != null) {
				for (int i = 0; i < netifDatas.length; i++) {
					INetworkInterfaceData netifData  = netifDatas[i];
					if (netifData != null) {
						NetworkInterface netif = new NetworkInterface(netifData);
						if (netif.equals(this)) {
							return i;
						}
					}
				}
			}
		} catch (SocketException e) {
			// purposely ignored, will return -1
		}
		return -1;
    }

    public String getDisplayName() {
		/* strict TCK conformance */
		return "".equals(displayName) ? null : displayName;
	}

	public static NetworkInterface getByName(String name) throws SocketException {
        if (name == null)
			throw new NullPointerException();

		INetworkInterfaceData[] netifDatas = VMNetworkInterface.getVMInterfaces();
		// specified to return null if no network interfaces
		if(netifDatas == null){
			return null;
		}

		for (int i = 0; i < netifDatas.length; i++) {
			INetworkInterfaceData netifData  = netifDatas[i];
			if ((netifData != null) && (name.equals(netifData.getName()))) {
				return new NetworkInterface(netifData);
			}

		}
		return null;
	}

    public static NetworkInterface getByIndex(int index) throws SocketException {
        if (index < 0) {
            throw new IllegalArgumentException("Interface index can't be negative");
        }
		INetworkInterfaceData[] netifDatas = VMNetworkInterface.getVMInterfaces();
		try {
			INetworkInterfaceData netifData  = netifDatas[index];
			if (netifData != null) {
				return new NetworkInterface(netifData);
			}
			return null;
		} catch(IndexOutOfBoundsException ex) {
			return null;
		}
    }

	public static NetworkInterface getByInetAddress(InetAddress addr) throws SocketException {
		if (addr == null) {
			throw new NullPointerException();
		}
		if (!(addr instanceof Inet4Address || addr instanceof Inet6Address)) {
			throw new IllegalArgumentException ("invalid address type");
		}
		INetworkInterfaceData[] netifDatas = VMNetworkInterface.getVMInterfaces();
		// specified to return null if no network interfaces
		if(netifDatas == null){
			return null;
		}

		NetworkInterface netif =  null;
		for (int i = 0; i < netifDatas.length; i++)
		{
			INetworkInterfaceData netifData  = netifDatas[i];
			byte[][] netifDataAddrs = null;

			if(netifData == null || (netifDataAddrs=netifData.getAddresses()) == null
					|| netifDataAddrs.length == 0){
				continue;
			}

			InetAddress[] netifAddrs = new InetAddress[netifDataAddrs.length];
			for(int j=0; j < netifDataAddrs.length; j++){
				InetAddress netifAddr;
				try {
					netifAddr = netifAddrs[j] = InetAddress.getByAddress(netifDataAddrs[j]);
				} catch (UnknownHostException e) {
					SocketException se = new SocketException();
					se.initCause(e);
					throw se;
				}

				if(netif == null && netifAddr.equals(addr)){
					netif = new NetworkInterface(netifData);
				}
			}
			if(netif != null){
				break;
			}
		}
		return netif;
	}


	public static Enumeration<NetworkInterface> getNetworkInterfaces()
			throws SocketException {


		INetworkInterfaceData[] netifDatas = VMNetworkInterface.getVMInterfaces();

		// specified to return null if no network interfaces
		if(netifDatas == null || netifDatas.length == 0){
			return null;
		}
		final ArrayList<NetworkInterface> netifss = new ArrayList<NetworkInterface>(netifDatas.length);

		for (int i = 0; i < netifDatas.length; i++)
		{
			INetworkInterfaceData netifData  = netifDatas[i];
			if(netifData == null ){
				continue;
			}

			NetworkInterface netif = new NetworkInterface(netifData);
			if(!netifss.contains(netif)){
				netifss.add(netif);
			}
		}

		if(netifss.isEmpty()){
			return null;
		}

		return new Enumeration<NetworkInterface>() {
			private int i = 0;
			public NetworkInterface nextElement() {
				if (i < netifss.size()) {
					NetworkInterface netif = netifss.get(i++);
					return netif;
				} else {
					throw new NoSuchElementException();
				}
			}

			public boolean hasMoreElements() {
				return (i < netifss.size());
			}
		};
	}


	public boolean isUp() throws SocketException {
		return VMNetworkInterface.isUp(name);
	}

	public boolean isLoopback() throws SocketException {
		return VMNetworkInterface.isLoopback(name);
	}


	public boolean supportsMulticast() throws SocketException {
		return VMNetworkInterface.supportsMulticast(name);
	}

	public byte[] getHardwareAddress() throws SocketException {
		return VMNetworkInterface.getHardwareAddress(name);
	}

	public int getMTU() throws SocketException {
		return VMNetworkInterface.getMTU(name);
	}

	public boolean equals(Object obj) {
		if (!(obj instanceof NetworkInterface)) {
			return false;
		}
		NetworkInterface that = (NetworkInterface)obj;
		if (this.name != null ) {
			if (!this.name.equals(that.name)) {
				return false;
			}
		} else {
			if (that.name != null) {
				return false;
			}
		}

		if (this.addrs == null) {
			return that.addrs == null;
		} else if (that.addrs == null) {
			return false;
		}

		/* Both addrs not null. Compare number of addresses */

		if (this.addrs.length != that.addrs.length) {
			return false;
		}

		InetAddress[] thatAddrs = that.addrs;
		int count = thatAddrs.length;

		for (int i=0; i<count; i++) {
			boolean found = false;
			for (int j=0; j<count; j++) {
				if (addrs[i].equals(thatAddrs[j])) {
					found = true;
					break;
				}
			}
			if (!found) {
				return false;
			}
		}
		return true;
	}

	public int hashCode() {
		return name == null? 0: name.hashCode();
	}

	public String toString() {
		String result = "name:";
		result += name == null? "null": name;
		if (displayName != null) {
			result += " (" + displayName + ")";
		}
		return result;
	}

}
