/*
 * Copyright (c) 1997, 2007, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package javax.net;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.SocketException;

public abstract class ServerSocketFactory {
	//
	// NOTE: JDK 1.1 bug in class GC, this can get collected
	// even though it's always accessible via getDefault().
	//
	private static ServerSocketFactory theFactory;

	protected ServerSocketFactory() {
	/* NOTHING */ }

	public static ServerSocketFactory getDefault() {
		synchronized (ServerSocketFactory.class) {
			if (theFactory == null) {
				//
				// Different implementations of this method could
				// work rather differently. For example, driving
				// this from a system property, or using a different
				// implementation than JavaSoft's.
				//
				theFactory = new DefaultServerSocketFactory();
			}
		}

		return theFactory;
	}

	public ServerSocket createServerSocket() throws IOException {
		throw new SocketException("Unbound server sockets not implemented");
	}

	public abstract ServerSocket createServerSocket(int port) throws IOException;

	public abstract ServerSocket createServerSocket(int port, int backlog) throws IOException;

	public abstract ServerSocket createServerSocket(int port, int backlog, InetAddress ifAddress) throws IOException;
}

//
// The default factory has NO intelligence. In fact it's not clear
// what sort of intelligence servers need; the onus is on clients,
// who have to know how to tunnel etc.
//
class DefaultServerSocketFactory extends ServerSocketFactory {

	DefaultServerSocketFactory() {
		/* NOTHING */
	}

	@Override
	public ServerSocket createServerSocket() throws IOException {
		return new ServerSocket();
	}

	@Override
	public ServerSocket createServerSocket(int port) throws IOException {
		return new ServerSocket(port);
	}

	@Override
	public ServerSocket createServerSocket(int port, int backlog) throws IOException {
		return new ServerSocket(port, backlog);
	}

	@Override
	public ServerSocket createServerSocket(int port, int backlog, InetAddress ifAddress) throws IOException {
		return new ServerSocket(port, backlog, ifAddress);
	}
}
