/*
 * Copyright (c) 1997, 2012, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */


package javax.net.ssl;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.SocketException;
import java.security.NoSuchAlgorithmException;

import javax.net.ServerSocketFactory;

/**
 * <code>SSLServerSocketFactory</code>s create
 * <code>SSLServerSocket</code>s.
 *
 * @since 1.4
 * @see SSLSocket
 * @see SSLServerSocket
 * @author David Brownell
 */
public abstract class SSLServerSocketFactory extends ServerSocketFactory
{
	private static SSLServerSocketFactory theFactory;

	private static boolean propertyChecked;

	private static void log(String msg) {
		if (SSLSocketFactory.DEBUG) {
			System.out.println(msg);
		}
	}

	/**
	 * Constructor is used only by subclasses.
	 */
	protected SSLServerSocketFactory() { /* NOTHING */ }

	public static synchronized ServerSocketFactory getDefault() {
		if (theFactory != null) {
			return theFactory;
		}

		if (propertyChecked == false) {
			propertyChecked = true;
			String clsName = SSLSocketFactory.getSecurityProperty
					("ssl.ServerSocketFactory.provider");
			if (clsName != null) {
				log("setting up default SSLServerSocketFactory");
				try {
					Class<?> cls = null;
					cls = Class.forName(clsName);
					log("class " + clsName + " is loaded");
					SSLServerSocketFactory fac = (SSLServerSocketFactory)cls.newInstance();
					log("instantiated an instance of class " + clsName);
					theFactory = fac;
					return fac;
				} catch (Exception e) {
					log("SSLServerSocketFactory instantiation failed: " + e);
					theFactory = new DefaultSSLServerSocketFactory(e);
					return theFactory;
				}
			}
		}

		try {
			return SSLContext.getDefault().getServerSocketFactory();
		} catch (NoSuchAlgorithmException e) {
			return new DefaultSSLServerSocketFactory(e);
		}
	}

}


//
// The default factory does NOTHING.
//
class DefaultSSLServerSocketFactory extends SSLServerSocketFactory {

	private final Exception reason;

	DefaultSSLServerSocketFactory(Exception reason) {
		this.reason = reason;
	}

	private ServerSocket throwException() throws SocketException {
		throw (SocketException)
		new SocketException(reason.toString()).initCause(reason);
	}

	@Override
	public ServerSocket createServerSocket() throws IOException {
		return throwException();
	}


	@Override
	public ServerSocket createServerSocket(int port)
			throws IOException
	{
		return throwException();
	}

	@Override
	public ServerSocket createServerSocket(int port, int backlog)
			throws IOException
	{
		return throwException();
	}

	@Override
	public ServerSocket
	createServerSocket(int port, int backlog, InetAddress ifAddress)
			throws IOException
	{
		return throwException();
	}

}
